"""
A Task is a unit of execution which can act on multiple repositories.

You define a task by instantiating a Task subclass and passing it to a Pipeline object.

Tasks are related to Repositories by Links. Links are created by
``Task.link("my_link_name", my_repository, **disposition)``. The main disposition kwargs you'll want to use are
``is_input`` and ``is_output``. See `Task.link` for more information.

For a shortcut for linking the output of one task as the input of another task, see `Task.plug`.

.. autodata:: STDOUT
"""
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Coroutine,
    Dict,
    Iterable,
    Optional,
    Protocol,
    Tuple,
    Union,
)
from abc import abstractmethod
from asyncio import Future
from concurrent.futures import FIRST_EXCEPTION, Executor, wait
from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
from enum import Enum, auto
from pathlib import Path
import asyncio
import inspect
import logging
import os
import sys
import traceback

from kubernetes_asyncio.client import ApiException, V1Pod
import aiofiles.os
import jinja2.async_utils
import jinja2.compiler
import yaml

from .pod_manager import PodManager
from .repository import (
    AggregateAndRepository,
    AggregateOrRepository,
    BlobRepository,
    BlockingRepository,
    ExecutorLiveRepo,
    LiveKubeRepository,
    MetadataRepository,
    RelatedItemRepository,
    Repository,
)
from .resource_manager import ResourceManager, Resources
from .utils import async_copyfile

if TYPE_CHECKING:
    from .proc_manager import AbstractProcessManager

l = logging.getLogger(__name__)

__all__ = (
    "Link",
    "Task",
    "KubeTask",
    "ProcessTask",
    "InProcessSyncTask",
    "ExecutorTask",
    "KubeFunctionTask",
    "STDOUT",
)

# pylint: disable=broad-except,bare-except


class RepoHandlingMode(Enum):
    """
    Mode for the `build_env` function.

    LAZY = Build repositories and links into repository references
    SMART = Build repositories and links into un-awaited `info` coroutines.
    EAGER = Build repositories and links into awaited `info` values.
    """

    LAZY = auto()
    SMART = auto()
    EAGER = auto()


async def build_env(env: Dict[str, Any], job: str, mode: RepoHandlingMode) -> Dict[str, Any]:
    """
    Given a mapping from environment key names to values that can be used during templating. The way repositories are
    transformed into job-related information is controlled by the `RepoHandlingMode`.
    """
    result = {}
    for key, val in env.items():
        if isinstance(val, Repository):
            repo = val
        elif isinstance(val, Link):
            repo = val.repo
        else:
            result[key] = val
            continue
        if mode == RepoHandlingMode.SMART:
            result[key] = repo
        elif mode == RepoHandlingMode.LAZY:
            result[key] = repo.info(job)
        else:
            result[key] = await repo.info(job)
    return result


async def render_template(template, env: Dict[str, Any]):
    """
    Given a template and an environment, use jinja2 to parameterize the template with the environment.

    :param template:  A template string, or a path to a local template file.
    :param env:       A mapping from environment key names to values.
    :return:          The rendered template, as a string.
    """
    j = jinja2.Environment(
        enable_async=True,
        keep_trailing_newline=True,
    )
    j.code_generator_class = ParanoidAsyncGenerator
    if await aiofiles.os.path.isfile(template):
        async with aiofiles.open(template, "r") as fp:
            template_str = await fp.read()
    else:
        template_str = template
    templating = j.from_string(template_str)
    return await templating.render_async(**env)


@dataclass
class Link:
    """
    The dataclass for holding linked repositories and their disposition metadata. Don't create these manually, instead
    use `Task.link`.
    """

    repo: Repository
    is_input: bool = False
    is_output: bool = False
    is_status: bool = False
    inhibits_start: bool = False
    required_for_start: bool = False
    inhibits_output: bool = False
    required_for_output: bool = False


class Task:
    """
    The Task base class.
    """

    def __init__(self, name: str, ready: Optional[Repository] = None):
        self.name = name
        self._ready = ready
        self.links: Dict[str, Link] = {}
        self.synchronous = False
        self.metadata = True

    def __repr__(self):
        return f"<{type(self).__name__} {self.name}>"

    @property
    def ready(self):
        """
        Return the repository whose job membership is used to determine whether a task instance should be launched.
        If an override is provided to the constructor, this is that, otherwise it is
        ``AND(*requred_for_start, NOT(OR(*inhibits_start)))``.
        """
        if self._ready is not None:
            return self._ready
        return BlockingRepository(
            AggregateAndRepository(**self.required_for_start),
            AggregateOrRepository(**self.inhibits_start),
        )

    def link(
        self,
        name: str,
        repo: Repository,
        is_input: bool = False,
        is_output: bool = False,
        is_status: bool = False,
        inhibits_start: bool = False,
        required_for_start: Optional[bool] = None,
        inhibits_output: bool = False,
        required_for_output: Optional[bool] = None,
    ):
        """
        Create a link between this task and a repository.

        :param name: The name of the link. Used only in the admin interface.
        :param repo: The repository to link.
        :param is_input: Whether this repository contains data which is consumed by the task. Default False.
        :param is_output: Whether this repository is populated by the task. Default False.
        :param is_status: Whether this task is populated with task-ephemeral data. Default False.
        :param inhibits_start: Whether membership in this repository should be used in the default ``ready`` repository
                               to prevent jobs for being launched. Default False.
        :param required_for_start: Whether membership in this repository should be used in the default ``ready``
                                   repository to allow jobs to be launched. If unspecified, defaults to ``is_input``.
        :param inhibits_output:     Whether this repository should become ``inhibits_start`` in tasks this task is
                                    plugged into. Default False.
        :param required_for_output: Whether this repository should become `required_for_start` in tasks this task is
                                    plugged into. If unspecified, defaults to ``is_output``.
        """
        if required_for_start is None:
            required_for_start = is_input
        if required_for_output is None:
            required_for_output = is_output

        self.links[name] = Link(
            repo=repo,
            is_input=is_input,
            is_output=is_output,
            is_status=is_status,
            inhibits_start=inhibits_start,
            required_for_start=required_for_start,
            inhibits_output=inhibits_output,
            required_for_output=required_for_output,
        )

    def plug(
        self,
        output: "Task",
        output_links: Optional[Iterable[str]] = None,
        meta: bool = True,
        translator: Optional[Repository] = None,
        translate_allow_deletes=False,
        translate_prefetch_lookup=True,
    ):
        """
        Link the output repositories from `output` as inputs to this task.

        :param output:  The task to plug into this one.
        :param output_links: Optional: An iterable allowlist of links to only use.
        :param meta: Whether to transfer repository inhibit- and required_for- dispositions. Default True.
        :param translator: Optional: A repository used to transform job identifiers. If this is provided, a job will be
                           said to be present in the resulting linked repository if the source repository contains the
                           key ``await translator.info(job)``.
        :param translate_allow_deletes:  If translator is provided, this controls whether attempts to delete from the
                                         translated repository will do anything. Default False.
        :param translate_prefetch_lookup: If translator is provided, this controls whether the translated repository
                                          will pre-fetch the list of translations on first access. Default True.
        """
        for name, link in output.links.items():
            link_attrs = {}
            if link.inhibits_output and meta:
                link_attrs["inhibits_start"] = True
            if link.is_output and (output_links is None or name in output_links):
                link_attrs["is_input"] = True
            if link.required_for_output and meta:
                link_attrs["required_for_start"] = True
            if not link.is_output:
                name = f"{output.name}_{name}"
            if link_attrs:
                repo = link.repo
                if translator is not None:
                    repo = RelatedItemRepository(
                        repo,
                        translator,
                        allow_deletes=translate_allow_deletes,
                        prefetch_lookup=translate_prefetch_lookup,
                    )
                self.link(name, repo, **link_attrs)

    @property
    def input(self):
        """
        A mapping from link name to repository for all links marked ``is_input``.
        """
        return {name: link.repo for name, link in self.links.items() if link.is_input}

    @property
    def output(self):
        """
        A mapping from link name to repository for all links marked ``is_output``.
        """
        return {name: link.repo for name, link in self.links.items() if link.is_output}

    @property
    def status(self):
        """
        A mapping from link name to repository for all links marked ``is_status``.
        """
        return {name: link.repo for name, link in self.links.items() if link.is_status}

    @property
    def inhibits_start(self):
        """
        A mapping from link name to repository for all links marked ``inhibits_start``.
        """
        return {name: link.repo for name, link in self.links.items() if link.inhibits_start}

    @property
    def required_for_start(self):
        """
        A mapping from link name to repository for all links marked ``required_for_start``.
        """
        return {name: link.repo for name, link in self.links.items() if link.required_for_start}

    @property
    def inhibits_output(self):
        """
        A mapping from link name to repository for all links marked ``inhibits_output``.
        """
        return {name: link.repo for name, link in self.links.items() if link.inhibits_output}

    @property
    def required_for_output(self):
        """
        A mapping from link name to repository for all links marked ``required_for_output``.
        """
        return {name: link.repo for name, link in self.links.items() if link.required_for_output}

    async def validate(self):
        """
        Raise an exception if for any reason the task is misconfigured. This is guaranteed to be called exactly once per
        pipeline, so it is safe to use for setup and initialization in an async context.
        """

    @abstractmethod
    async def update(self) -> bool:
        """
        Part one of the pipeline maintenance loop. Override this to perform any maintenance operations on the set of
        live tasks. Typically, this entails reaping finished processes.

        Returns True if literally anything interesting happened, or if there are any live tasks.
        """
        raise NotImplementedError

    async def launch_all(self):
        """
        Part two of the pipeline maintenance loop. Do not override this; the default implementation is typically pretty
        good - it enumerates `ready` and calls `launch` for each ready job.

        Returns True if there were any jobs to launch.
        """
        launchers = [self._launch(job) async for job in self.ready]
        await asyncio.gather(*launchers)
        return bool(launchers)

    async def _launch(self, job):
        try:
            l.debug("Launching %s:%s", self.name, job)
            await self.launch(job)
        except:
            l.exception("Failed to launch %s:%s", self, job)

    @abstractmethod
    async def launch(self, job):
        """
        Launch a job. Override this to begin execution of the provided job. Error handling will be done for you.
        """
        raise NotImplementedError


class ParanoidAsyncGenerator(jinja2.compiler.CodeGenerator):
    """
    A class to instrument jinja2 to be more aggressive about awaiting objects and to cache their results. Probably
    don't use this directly.
    """

    def write_commons(self):
        self.writeline("from jinja2.async_utils import _common_primitives")
        self.writeline("import inspect")
        self.writeline("seen = {}")
        self.writeline("async def auto_await2(value):")
        self.writeline("    if type(value) in _common_primitives:")
        self.writeline("        return value")
        self.writeline("    if inspect.isawaitable(value):")
        self.writeline("        cached = seen.get(value)")
        self.writeline("        if cached is None:")
        self.writeline("            cached = await value")
        self.writeline("            seen[value] = cached")
        self.writeline("        return cached")
        self.writeline("    return value")
        super().write_commons()

    def visit_Name(self, node, frame):
        if self.environment.is_async:
            self.write("(await auto_await2(")

        super().visit_Name(node, frame)

        if self.environment.is_async:
            self.write("))")


class KubeTask(Task):
    """
    A task which runs a kubernetes pod.

    Will automatically link a `LiveKubeRepository` as "live" with
    ``inhibits_start, inhibits_output, is_status``
    """

    def __init__(
        self,
        name: str,
        podman: Callable[[], PodManager],
        resources: ResourceManager,
        template: Union[str, Path],
        logs: Optional[BlobRepository],
        done: Optional[MetadataRepository],
        timeout: Optional[timedelta] = None,
        env: Optional[Dict[str, Any]] = None,
        ready: Optional[Repository] = None,
    ):
        """
        :param name: The name of the task.
        :param podman: A callable returning a PodManager to use to connect to the cluster.
        :param resources: A ResourceManager instance. Tasks launched will contribute to its quota and be denied if they
                          would break the quota.
        :param template: YAML markup for a pod manifest template, either as a string or a path to a file.
        :param logs: Optional: A BlobRepository to dump pod logs to on completion. Linked as "logs" with
                               ``inhibits_start, required_for_output, is_status``.
        :param done: A MetadataRepository in which to dump some information about the pod's lifetime and termination on
                     completion. Linked as "done" with ``inhibits_start, required_for_output, is_status``.
        :param timeout: Optional: When a pod is found to have been running continuously for this amount of time, it will
                        be timed out and stopped. The method `handle_timeout` will be called in-process.
        :param env:     Optional: Additional keys to add to the template environment.
        :param ready:   Optional: A repository from which to read task-ready status.

        It is highly recommended to provide one or more of ``done`` or ``logs`` so that at least one link is present
        with ``inhibits_start``.
        """
        super().__init__(name, ready)

        self.template = template
        self.resources = resources
        self._podman = podman
        self.logs = logs
        self.timeout = timeout
        self.done = done
        self.env = env if env is not None else {}
        self.warned = False

        self.resources.register(self._get_load)

        self.link(
            "live",
            LiveKubeRepository(self),
            is_status=True,
            inhibits_start=True,
            inhibits_output=True,
        )
        if logs:
            self.link(
                "logs",
                logs,
                is_status=True,
                inhibits_start=True,
                required_for_output=True,
            )
        if done:
            self.link(
                "done",
                done,
                is_status=True,
                inhibits_start=True,
                required_for_output=True,
            )

    @property
    def podman(self) -> PodManager:
        """
        The pod manager instance for this task. Will raise an error if the manager is provided by an unopened session.
        """
        return self._podman()

    async def _get_load(self):
        usage = Resources()

        try:
            for pod in (
                await self.podman.v1.list_namespaced_pod(
                    self.podman.namespace, label_selector=f"app={self.podman.app},task={self.name}"
                )
            ).items:
                for container in pod.spec.containers:
                    usage += Resources.parse(
                        container.resources.requests["cpu"], container.resources.requests["memory"]
                    )
        except ApiException as e:
            if e.reason != "Forbidden":
                raise

        return usage

    async def launch(self, job):
        env_input = dict(vars(self))
        env_input.update(self.links)
        env_input.update(self.env)
        env = await build_env(env_input, job, RepoHandlingMode.LAZY)
        env["job"] = job
        env["task"] = self.name
        env["argv0"] = os.path.basename(sys.argv[0])
        manifest = yaml.safe_load(await render_template(self.template, env))
        for item in env.values():
            if asyncio.iscoroutine(item):
                item.close()

        assert manifest["kind"] == "Pod"
        spec = manifest["spec"]
        spec["restartPolicy"] = "Never"

        request = Resources()
        for container in spec["containers"]:
            request += Resources.parse(
                container["resources"]["requests"]["cpu"], container["resources"]["requests"]["memory"]
            )

        limit = await self.resources.reserve(request)
        if limit is None:
            try:
                await self.podman.launch(job, self.name, manifest)
            except:
                await self.resources.relinquish(request)
                raise
        elif not self.warned:
            l.warning("Cannot launch %s: %s limit", self, limit)
            self.warned = True

    async def _cleanup(self, pod: V1Pod, reason: str):
        job = pod.metadata.labels["job"]

        if self.logs is not None:
            async with await self.logs.open(job, "w") as fp:
                try:
                    await fp.write(await self.podman.logs(pod))
                except TimeoutError:
                    await fp.write("<failed to fetch logs>\n")
        if self.done is not None:
            data = {
                "reason": reason,
                "start_time": pod.metadata.creation_timestamp,
                "end_time": datetime.now(timezone.utc),
                "image": pod.status.container_statuses[0].image,
                "node": pod.spec.node_name,
            }
            await self.done.dump(job, data)

        await self.delete(pod)

    async def delete(self, pod: V1Pod):
        """
        Kill a pod and relinquish its resources without marking the task as complete.
        """
        request = Resources()
        for container in pod.spec.containers:
            request += Resources.parse(container.resources.requests["cpu"], container.resources.requests["memory"])
        await self.podman.delete(pod)
        await self.resources.relinquish(request)

    async def update(self):
        self.warned = False
        result = False

        pods = await self.podman.query(task=self.name)
        for pod in pods:
            result = True
            try:
                uptime: timedelta = datetime.now(timezone.utc) - pod.metadata.creation_timestamp
                total_min = uptime.total_seconds() // 60
                uptime_hours, uptime_min = divmod(total_min, 60)
                l.debug(
                    "Pod %s is alive for %dh%dm",
                    pod.metadata.name,
                    uptime_hours,
                    uptime_min,
                )
                if pod.status.phase in ("Succeeded", "Failed"):
                    l.debug("...finished: %s", pod.status.phase)
                    await self._cleanup(pod, pod.status.phase)
                elif self.timeout is not None and uptime > self.timeout:
                    l.debug("...timed out")
                    await self.handle_timeout(pod)
                    await self._cleanup(pod, "Timeout")
            except Exception:
                l.exception("Failed to update kube task %s:%s", self.name, pod.metadata.name)
        return result

    async def handle_timeout(self, pod: V1Pod):
        """
        You may override this method in a subclass, and it will be called whenever a pod times out. You can use this
        method to e.g. scrape in-progress data out of the pod via an exec.
        """


class _StderrIsStdout:
    pass


STDOUT = _StderrIsStdout()


class ProcessTask(Task):
    """
    A task that runs a script. The interpreter is specified by the shebang, or the default shell if none present.
    The execution environment for the task is defined by the ProcessManager instance provided as an argument.
    """

    def __init__(
        self,
        name: str,
        manager: Callable[[], "AbstractProcessManager"],
        resource_manager: "ResourceManager",
        job_resources: "Resources",
        pids: MetadataRepository,
        template: str,
        environ: Optional[Dict[str, str]] = None,
        done: Optional[MetadataRepository] = None,
        stdin: Optional[BlobRepository] = None,
        stdout: Optional[BlobRepository] = None,
        stderr: Optional[Union[BlobRepository, _StderrIsStdout]] = None,
        ready: Optional[Repository] = None,
    ):
        """
        :param name: The name of this task.
        :param manager: A callable returnint the process manager with control over the target execution environment.
        :param resource_manager: A ResourceManager instance. Tasks launched will contribute to its quota and be denied
                                 if they would break the quota.
        :param job_resources: The amount of resources an individual job should contribute to the quota. Note that this
                              is currently **not enforced** target-side, so jobs may actually take up more resources
                              than assigned.
        :param pids: A metadata repository used to store the current live-status of processes. Will automatically be
                     linked as "pids" with ``is_status, inhibits_start, inhibits_output``.
        :param template: YAML markup for the template of a script to run, either as a string or a path to a file.
        :param environ: Additional environment variables to set on the target machine before running the task.
        :param done: Optional: A metadata repository in which to dump some information about the process's lifetime and
                     termination on completion. Linked as "done" with
                     ``inhibits_start, required_for_output, is_status``.
        :param stdin: Optional: A blob repository from which to source the process' standard input. The content will be
                      preloaded and transferred to the target environment, so the target does not need to be
                      authenticated to this repository. Linked as "stdin" with ``is_input``.
        :param stdout: Optional: A blob repository into which to dump the process' standard output. The content will be
                       transferred from the target environment on completion, so the target does not need to be
                       authenticated to this repository. Linked as "stdout" with ``is_output``.
        :param stderr: Optional: A blob repository into which to dump the process' standard error, or the constant
                       `pydatatask.task.STDOUT` to indicate that the stream should be interleaved with stdout.
                       Otherwise, the content will be transferred from the target environment on completion, so the
                       target does not need to be authenticated to this repository. Linked as "stderr" with
                       ``is_output``.
        :param ready:  Optional: A repository from which to read task-ready status.

        It is highly recommended to provide at least one of ``done``, ``stdout``, or ``stderr``, so that at least one
        link is present with ``inhibits_start``.
        """
        super().__init__(name, ready=ready)

        self.pids = pids
        self.template = template
        self.environ = environ
        self.done = done
        self.stdin = stdin
        self.stdout = stdout
        self._stderr = stderr
        self.job_resources = job_resources
        self.resource_manager = resource_manager
        self._manager = manager
        self.warned = False

        self.resource_manager.register(self._get_load)

        self.link("pids", pids, is_status=True, inhibits_start=True, inhibits_output=True)
        if stdin is not None:
            self.link("stdin", stdin, is_input=True)
        if stdout is not None:
            self.link("stdout", stdout, is_output=True)
        if isinstance(stderr, BlobRepository):
            self.link("stderr", stderr, is_output=True)
        if done is not None:
            self.link("done", done, is_status=True, required_for_output=True, inhibits_start=True)

    @property
    def manager(self) -> "AbstractProcessManager":
        """
        The process manager for this task. Will raise an error if the manager comes from a session which is closed.
        """
        return self._manager()

    async def _get_load(self) -> "Resources":
        count = 0
        async for _ in self.pids:
            count += 1
        return self.job_resources * count

    @property
    def stderr(self) -> Optional[BlobRepository]:
        """
        The repository into which stderr will be dumped, or None if it will go to the null device.
        """
        if isinstance(self._stderr, BlobRepository):
            return self._stderr
        else:
            return self.stdout

    @property
    def _unique_stderr(self):
        return self._stderr is not None and not isinstance(self._stderr, _StderrIsStdout)

    @property
    def basedir(self) -> Path:
        """
        The path in the target environment that will be used to store information about this task.
        """
        return self.manager.basedir / self.name

    async def update(self):
        self.warned = False
        job_map = await self.pids.info_all()
        pid_map = {meta["pid"]: job for job, meta in job_map.items()}
        expected_live = set(pid_map)
        try:
            live_pids = await self.manager.get_live_pids(expected_live)
        except Exception:
            l.error("Could not load live PIDs for %s", self, exc_info=True)
        else:
            died = expected_live - live_pids
            for pid in died:
                job = pid_map[pid]
                start_time = job_map[job]["start_time"]
                await self._reap(job, start_time)
        return bool(expected_live)

    async def launch(self, job):
        limit = await self.resource_manager.reserve(self.job_resources)
        if limit is not None:
            if not self.warned:
                l.warning("Cannot launch %s: %s limit", self, limit)
                self.warned = True
            return

        pid = None
        dirmade = False

        try:
            cwd = self.basedir / job / "cwd"
            await self.manager.mkdir(cwd)  # implicitly creates basedir / task / job
            dirmade = True
            stdin = None
            if self.stdin is not None:
                stdin = self.basedir / job / "stdin"
                async with await self.stdin.open(job, "rb") as fpr, await self.manager.open(stdin, "wb") as fpw:
                    await async_copyfile(fpr, fpw)
                stdin = str(stdin)
            stdout = None if self.stdout is None else str(self.basedir / job / "stdout")
            stderr = STDOUT
            if not isinstance(self._stderr, _StderrIsStdout):
                stderr = None if self._stderr is None else str(self.basedir / job / "stderr")
            env_src = dict(self.links)
            env_src["job"] = job
            env_src["task"] = self.name
            env = await build_env(env_src, job, RepoHandlingMode.LAZY)
            exe_path = self.basedir / job / "exe"
            exe_txt = await render_template(self.template, env)
            for item in env.values():
                if asyncio.iscoroutine(item):
                    item.close()
            async with await self.manager.open(exe_path, "w") as fp:
                await fp.write(exe_txt)
            pid = await self.manager.spawn(
                [str(exe_path)], self.environ, str(cwd), str(self.basedir / job / "return_code"), stdin, stdout, stderr
            )
            if pid is not None:
                await self.pids.dump(job, {"pid": pid, "start_time": datetime.now(tz=timezone.utc)})
        except:  # CLEAN UP YOUR MESS
            try:
                await self.resource_manager.relinquish(self.job_resources)
            except:
                pass
            try:
                if pid is not None:
                    await self.manager.kill(pid)
            except:
                pass
            try:
                if dirmade:
                    await self.manager.rmtree(self.basedir / job)
            except:
                pass
            raise

    async def _reap(self, job: str, start_time: datetime):
        try:
            if self.stdout is not None:
                async with await self.manager.open(self.basedir / job / "stdout", "rb") as fpr, await self.stdout.open(
                    job, "wb"
                ) as fpw:
                    await async_copyfile(fpr, fpw)
            if isinstance(self._stderr, BlobRepository):
                async with await self.manager.open(self.basedir / job / "stderr", "rb") as fpr, await self._stderr.open(
                    job, "wb"
                ) as fpw:
                    await async_copyfile(fpr, fpw)
            if self.done is not None:
                async with await self.manager.open(self.basedir / job / "return_code", "r") as fp1:
                    code = int(await fp1.read())
                await self.done.dump(
                    job, {"return_code": code, "start_time": start_time, "end_time": datetime.now(tz=timezone.utc)}
                )
            await self.manager.rmtree(self.basedir / job)
            await self.pids.delete(job)
            await self.resource_manager.relinquish(self.job_resources)
        except Exception:
            l.error("Could not reap process for %s:%s", self, job, exc_info=True)


class FunctionTaskProtocol(Protocol):
    """
    The protocol which is expected by the tasks which take a python function to execute.
    """

    def __call__(self, job: str, **kwargs) -> Coroutine:
        ...


class InProcessSyncTask(Task):
    """
    A task which runs in-process. Typical usage of this task might look like the following:

    .. code:: python

        @pydatatask.InProcessSyncTask("my_task", done_repo)
        async def my_task(job: str, inp: pydatatask.MetadataRepository, out: pydatatask.MetadataRepository):
            await out.dump(job, await inp.info(job))

        my_task.link("inp", repo_input, is_input=True)
        my_task.link("out", repo_output, is_output=True)
    """

    def __init__(
        self,
        name: str,
        done: MetadataRepository,
        ready: Optional[Repository] = None,
        func: Optional[FunctionTaskProtocol] = None,
    ):
        """
        :param name: The name of this task.
        :param done: A metadata repository to store some information about a job's runtime and termination on
                     completion.
        :param ready: Optional: A repository from which to read task-ready status.
        :param func: Optional: The async function to run as the task body, if you don't want to use this task as a
                     decorator.
        """
        super().__init__(name, ready=ready)

        self.done = done
        self.func = func
        self.link("done", done, is_status=True, inhibits_start=True, required_for_output=True)
        self._env: Dict[str, Any] = {}

    def __call__(self, f: FunctionTaskProtocol) -> "InProcessSyncTask":
        self.func = f
        return self

    async def validate(self):
        if self.func is None:
            raise ValueError("InProcessSyncTask.func is None")

        sig = inspect.signature(self.func, follow_wrapped=True)
        for name in sig.parameters.keys():
            if name == "job":
                self._env[name] = None
            elif name in self.links:
                self._env[name] = self.links[name].repo
            else:
                raise NameError("%s takes parameter %s but no such argument is available" % (self.func, name))

    async def update(self):
        pass

    async def launch(self, job):
        start_time = datetime.now()
        l.debug("Launching in-process %s:%s...", self.name, job)
        args = dict(self._env)
        if "job" in args:
            args["job"] = job
        args = await build_env(args, job, RepoHandlingMode.SMART)
        try:
            await self.func(**args)
        except Exception as e:
            l.info("In-process task %s:%s failed", self.name, job, exc_info=True)
            result = {
                "result": "exception",
                "exception": repr(e),
                "traceback": traceback.format_tb(e.__traceback__),
            }
        else:
            l.debug("...success")
            result = {"result": "success"}
        result["start_time"] = start_time
        result["end_time"] = datetime.now()
        if self.metadata:
            await self.done.dump(job, result)


class ExecutorTask(Task):
    """
    A task which runs python functions in a :external:class:`concurrent.futures.Executor`. This has not been tested on
    anything but the :external:class:`concurrent.futures.ThreadPoolExecutor`, so beware!

    See `InProcessSyncTask` for information on how to use instances of this class as decorators for their bodies.

    It is expected that the executor will perform all necessary resource quota management.
    """

    def __init__(
        self,
        name: str,
        executor: Executor,
        done: MetadataRepository,
        ready: Optional[Repository] = None,
        func: Optional[Callable] = None,
    ):
        """
        :param name: The name of this task.
        :param executor: The executor to run jobs in.
        :param done: A metadata repository to store some information about a job's runtime and termination on
                     completion.
        :param ready: Optional: A repository from which to read task-ready status.
        :param func: Optional: The async function to run as the task body, if you don't want to use this task as a
                     decorator.
        """
        super().__init__(name, ready)

        self.executor = executor
        self.func = func
        self.jobs: Dict[Future, Tuple[str, datetime]] = {}
        self.rev_jobs: Dict[str, Future] = {}
        self.live = ExecutorLiveRepo(self)
        self.done = done
        self._env: Dict[str, Any] = {}
        self.link("live", self.live, is_status=True, inhibits_output=True, inhibits_start=True)
        self.link(
            "done",
            self.done,
            is_status=True,
            required_for_output=True,
            inhibits_start=True,
        )

    def __call__(self, f: Callable) -> "ExecutorTask":
        self.func = f
        return self

    async def update(self):
        result = bool(self.jobs)
        done, _ = wait(self.jobs, 0, FIRST_EXCEPTION)
        for finished_job in done:
            job, start_time = self.jobs.pop(finished_job)
            # noinspection PyAsyncCall
            self.rev_jobs.pop(job)
            await self._cleanup(finished_job, job, start_time)
        return result

    async def _cleanup(self, job_future, job, start_time):
        e = job_future.exception()
        if e is not None:
            l.info("Executor task %s:%s failed", self.name, job, exc_info=e)
            data = {
                "result": "exception",
                "exception": repr(e),
                "traceback": traceback.format_tb(e.__traceback__),
                "end_time": datetime.now(),
            }
        else:
            l.debug("...executor task %s:%s success", self.name, job)
            data = {"result": "success", "end_time": job_future.result()}
        data["start_time"] = start_time
        await self.done.dump(job, data)

    async def validate(self):
        if self.func is None:
            raise ValueError("InProcessAsyncTask %s has func None" % self.name)

        sig = inspect.signature(self.func, follow_wrapped=True)
        for name in sig.parameters.keys():
            if name == "job":
                self._env[name] = None
            elif name in self.links:
                self._env[name] = self.links[name].repo
            else:
                raise NameError("%s takes parameter %s but no such argument is available" % (self.func, name))

    async def launch(self, job):
        l.debug("Launching %s:%s with %s...", self.name, job, self.executor)
        args = dict(self._env)
        if "job" in args:
            args["job"] = job
        args = await build_env(args, job, RepoHandlingMode.SMART)
        start_time = datetime.now()
        running_job = self.executor.submit(self._timestamped_func, self.func, args)
        if self.synchronous:
            while not running_job.done():
                await asyncio.sleep(0.1)
            await self._cleanup(running_job, job, start_time)
        else:
            self.jobs[running_job] = (job, start_time)
            self.rev_jobs[job] = running_job

    async def cancel(self, job):
        """
        Stop the current job from running, or do nothing if it is not running.
        """
        future = self.rev_jobs.pop(job)
        if future is not None:
            future.cancel()
            self.jobs.pop(future)

    @staticmethod
    def _timestamped_func(func, args):
        loop = asyncio.new_event_loop()
        loop.run_until_complete(func(**args))
        return datetime.now()


class KubeFunctionTask(KubeTask):
    """
    A task which runs a python function on a kubernetes cluster. Requires a pod template which will execute a python
    script calling `pydatatask.main.main`. This works by running ``python3 main.py launch [task] [job] --sync``.

    Sample usage:

    .. code:: python

        @KubeFunctionTask(
            "my_task",
            podman,
            resman,
            '''
                apiVersion: v1
                kind: Pod
                spec:
                  containers:
                    - name: leader
                      image: "docker.example.com/my/image"
                      command:
                        - python3
                        - {{argv0}}
                        - launch
                        - "{{task}}"
                        - "{{job}}"
                        - "--force"
                        - "--sync"
                      resources:
                        requests:
                          cpu: 100m
                          memory: 256Mi
            ''',
            repo_logs,
            repo_done,
            repo_func_done,
        )
        async def my_task(job: str, inp: pydatatask.MetadataRepository, out: pydatatask.MetadataRepository):
            await out.dump(job, await inp.info(job))

        my_task.link("inp", repo_input, is_input=True)
        my_task.link("out", repo_output, is_output=True)
    """

    def __init__(
        self,
        name: str,
        podman: Callable[[], PodManager],
        resources: ResourceManager,
        template: Union[str, Path],
        logs: Optional[BlobRepository] = None,
        kube_done: Optional[MetadataRepository] = None,
        func_done: Optional[MetadataRepository] = None,
        env: Optional[Dict[str, Any]] = None,
        func: Optional[Callable] = None,
    ):
        """
        :param name: The name of this task.
        :param podman: A callable returning a PodManager to use to connect to the cluster.
        :param resources: A ResourceManager instance. Tasks launched will contribute to its quota and be denied if they
                          would break the quota.
        :param template: YAML markup for a pod manifest template that will run `pydatatask.main.main` as
                         ``python3 main.py launch [task] [job] --sync --force``, either as a string or a path to a file.
        :param logs: Optional: A BlobRepository to dump pod logs to on completion. Linked as "logs" with
                               ``inhibits_start, required_for_output, is_status``.
        :param kube_done: Optional: A MetadataRepository in which to dump some information about the pod's lifetime and
                          termination on completion. Linked as "done" with
                          ``inhibits_start, required_for_output, is_status``.
        :param func_done: Optional: A MetadataRepository in which to dump some information about the function's lifetime
                          and termination on completion. Linked as "func_done" with
                          ``inhibits_start, required_for_output, is_status``.
        :param env:     Optional: Additional keys to add to the template environment.
        :param ready:   Optional: A repository from which to read task-ready status.
        :param func: Optional: The async function to run as the task body, if you don't want to use this task as a
                     decorator.

        It is highly recommended to provide at least one of ``kube_done``, ``func_done``, or ``logs``, so that at least
        one link is present with ``inhibits_start``.
        """
        super().__init__(name, podman, resources, template, logs, kube_done, env=env)
        self.func = func
        self.func_done = func_done
        if func_done is not None:
            self.link(
                "func_done",
                func_done,
                required_for_output=True,
                is_status=True,
                inhibits_start=True,
            )
        self._func_env: Dict[str, Any] = {}

    def __call__(self, f: Callable) -> "KubeFunctionTask":
        self.func = f
        return self

    async def validate(self):
        if self.func is None:
            raise ValueError("KubeFunctionTask %s has func None" % self.name)

        sig = inspect.signature(self.func, follow_wrapped=True)
        for name in sig.parameters.keys():
            if name == "job":
                self._func_env[name] = None
            elif name in self.links:
                self._func_env[name] = self.links[name].repo
            else:
                raise NameError("%s takes parameter %s but no such argument is available" % (self.func, name))

    async def launch(self, job):
        if self.synchronous:
            await self._launch_sync(job)
        else:
            await super().launch(job)

    async def _launch_sync(self, job):
        start_time = datetime.now()
        l.debug("Launching --sync %s:%s...", self.name, job)
        args = dict(self._func_env)
        if "job" in args:
            args["job"] = job
        args = await build_env(args, job, RepoHandlingMode.SMART)
        try:
            await self.func(**args)
        except Exception as e:
            l.info("--sync task %s:%s failed", self.name, job, exc_info=True)
            result = {
                "result": "exception",
                "exception": repr(e),
                "traceback": traceback.format_tb(e.__traceback__),
            }
        else:
            l.debug("...success")
            result = {"result": "success"}
        result["start_time"] = start_time
        result["end_time"] = datetime.now()
        if self.metadata:
            await self.func_done.dump(job, result)



"""
A session is a tool for managing multiple live resources. Async resource manager routines can be registered, and in
their place will be left a callable which will return the live resource while the session is opened.

For example:

.. code:: python

    session = pydatatask.Session()

    @session.resource
    async def my_file_dst():
        with open('/tmp/my_file_dst', 'wb') as fp:
            yield fp

    @session.resource
    async def my_file_src():
        with open('/tmp/my_file_src', 'rb') as fp:
            yield fp

    async with session:
        my_file_dst().write(my_file_src().read())

The advantage over using normal contextlib context managers is that this produces a function reference which can be
passed into other locations in sync contexts with the promise that it will not be called until the session is opened.

If a session is passed to a pipeline, it will be opened and closed when the pipeline is opened and closed.

Sessions cannot be opened more than once. But this doesn't have to be the way! If you have a use case, complain in a
GitHub issue, and I'll see what can be done.
"""
from typing import AsyncGenerator, Callable, TypeVar

__all__ = ("Session",)

T = TypeVar("T")


class Session:
    """
    The session class. See module docs for usage information.
    """

    def __init__(self):
        self._resource_defs = {}
        self.resources = {}

    def resource(self, manager: Callable[[], AsyncGenerator[T, None]]) -> Callable[[], T]:
        """
        Decorator for resource managers. Should be called with an async function that will yield exactly one object, the
        live constructed resource, and then tear that resource down on completion.
        """
        self._resource_defs[manager.__name__] = manager()

        def inner():
            if manager.__name__ not in self.resources:
                raise Exception("Session is not open")
            return self.resources[manager.__name__]

        return inner

    async def __aenter__(self):
        await self.open()

    async def open(self):
        """
        Open the session, initializing all the resources.

        This is automatically called when entering an ``async with session:`` block.
        """
        for name, manager in self._resource_defs.items():
            self.resources[name] = await manager.__anext__()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        """
        Close the session, tearing down all the resources.

        This is automatically called when exiting an ``async with session:`` block.
        """
        for name, manager in self._resource_defs.items():
            try:
                await manager.__anext__()
            except StopAsyncIteration:
                pass
            else:
                print("Warning: resource has more than one yield")
            self.resources.pop(name)



"""
In order for a `KubeTask` or a subclass to connect, authenticate, and manage pods in a kubernetes cluster, it needs
several resource references. the `PodManager` simplifies tracking the lifetimes of these resources.
"""

from typing import Callable, List, Optional
import logging

from kubernetes_asyncio.client import ApiClient, CoreV1Api, V1Pod
from kubernetes_asyncio.config.kube_config import Configuration
from kubernetes_asyncio.stream import WsApiClient

l = logging.getLogger(__name__)

__all__ = ("PodManager",)


class PodManager:
    """
    A pod manager allows multiple tasks to share a connection to a kubernetes cluster and manage pods on it.
    """

    def __init__(
        self,
        app: str,
        namespace: str,
        config: Optional[Callable[[], Configuration]] = None,
    ):
        """
        :param app: The app name string with which to label all created pods.
        :param namespace: The namespace in which to create and query pods.
        :param config: Optional: A callable returning a kubernetes configuration object. If not provided, will attempt
                                 to use the "default" configuration, i.e. what is available after calling
                                 ``await kubernetes_asyncio.config.load_kube_config()``.
        """
        self.app = app
        self.namespace = namespace
        self._config = config

        self._api: Optional[ApiClient] = None
        self._api_ws: Optional[WsApiClient] = None
        self._v1 = None
        self._v1_ws = None

    @property
    def api(self):
        """
        The current API client.
        """
        if self._api is None:
            if self._config is None:
                self._api = ApiClient()
            else:
                self._api = ApiClient(self._config())
        return self._api

    @property
    def api_ws(self) -> WsApiClient:
        """
        The current websocket-aware API client.
        """
        if self._api_ws is None:
            if self._config is None:
                self._api_ws = WsApiClient()
            else:
                self._api_ws = WsApiClient(self._config())
        return self._api_ws

    @property
    def v1(self) -> CoreV1Api:
        """
        A CoreV1Api instance associated with the current API client.
        """
        if self._v1 is None:
            self._v1 = CoreV1Api(self.api)
        return self._v1

    @property
    def v1_ws(self) -> CoreV1Api:
        """
        A CoreV1Api instance associated with the current websocket-aware API client.
        """
        if self._v1_ws is None:
            self._v1_ws = CoreV1Api(self.api_ws)
        return self._v1_ws

    async def close(self):
        """
        Close the network connections associated with this podman.
        """
        if self._api is not None:
            await self._api.close()
            self._api = None
        if self._api_ws is not None:
            await self._api_ws.close()
            self._api_ws = None

    async def launch(self, job, task, manifest):
        """
        Create a pod with the given manifest, named and labeled for this podman's app and the given job and task.
        """
        assert manifest["kind"] == "Pod"

        manifest["metadata"] = manifest.get("metadata", {})
        manifest["metadata"].update(
            {
                "name": "%s-%s-%s" % (self.app, job, task),
                "labels": {
                    "app": self.app,
                    "task": task,
                    "job": job,
                },
            }
        )

        l.info("Creating task %s for job %s", task, job)
        await self.v1.create_namespaced_pod(self.namespace, manifest)

    async def query(self, job=None, task=None) -> List[V1Pod]:
        """
        Return a list of pods labeled for this podman's app and (optional) the given job and task.
        """
        selectors = ["app=" + self.app]
        if job is not None:
            selectors.append("job=" + job)
        if task is not None:
            selectors.append("task=" + task)
        selector = ",".join(selectors)
        return (await self.v1.list_namespaced_pod(self.namespace, label_selector=selector)).items

    async def delete(self, pod: V1Pod):
        """
        Destroy the given pod.
        """
        await self.v1.delete_namespaced_pod(pod.metadata.name, self.namespace)

    async def logs(self, pod: V1Pod, timeout=10) -> str:
        """
        Retrieve the logs for the given pod.
        """
        return await self.v1.read_namespaced_pod_log(pod.metadata.name, self.namespace, _request_timeout=timeout)



"""
Documentation for this module can be found at https://pydatatask.readthedocs.io/en/stable/
"""

from .repository import *
from .task import *
from .pipeline import *
from .session import *
from .main import *
from .resource_manager import *
from .pod_manager import *
from .proc_manager import *

__version__ = "0.2.2"
released_version = "0.2.2"



"""
Various utility classes and functions that are used throughout the codebase but don't belong anywhere in particular.
"""

from typing import AsyncContextManager, List, Optional, Protocol
import codecs


class AReadStream(Protocol, AsyncContextManager):
    """
    A protocol for reading data from an asynchronous stream.
    """

    async def read(self, n: Optional[int] = None) -> bytes:
        """
        Read and return up to ``n`` bytes, or if unspecified, the rest of the stream.
        """

    async def close(self) -> None:
        """
        Close and release the stream.
        """


class AWriteStream(Protocol, AsyncContextManager):
    """
    A protocol for writing data to an asynchronous stream.
    """

    async def write(self, data: bytes):
        """
        Write ``data`` to the stream.
        """

    async def close(self) -> None:
        """
        Close and release the stream.
        """


async def async_copyfile(copyfrom: AReadStream, copyto: AWriteStream, blocksize=1024 * 1024):
    """
    Stream data from ``copyfrom`` to ``copyto``.
    """
    while True:
        data = await copyfrom.read(blocksize)
        if not data:
            break
        await copyto.write(data)


class AReadText:
    """
    An async version of :external:class:`io.TextIOWrapper` which can only handle reading.
    """

    def __init__(
        self,
        base: AReadStream,
        encoding: str = "utf-8",
        errors="strict",
        chunksize=4096,
    ):
        self.base = base
        self.decoder = codecs.lookup(encoding).incrementaldecoder(errors)
        self.buffer = ""
        self.chunksize = chunksize

    async def read(self, n: Optional[int] = None) -> str:
        """
        Read up to ``n`` chars from the string, or the rest of the stream if not provided.
        """
        while n is None or len(self.buffer) < n:
            data = await self.base.read(self.chunksize)
            self.buffer += self.decoder.decode(data, final=not bool(data))
            if not data:
                break

        if n is not None:
            result, self.buffer = self.buffer[:n], self.buffer[:n]
        else:
            result, self.buffer = self.buffer, ""
        return result

    async def close(self):
        """
        Close and release the stream.
        """
        await self.base.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class AWriteText:
    """
    An async version of ``io.TextIOWrapper`` which can only handle writing.
    """

    def __init__(self, base: AWriteStream, encoding="utf-8", errors="strict"):
        self.base = base
        self.encoding = encoding
        self.errors = errors

    async def write(self, data: str):
        """
        Write ``data`` to the stream.
        """
        await self.base.write(data.encode(self.encoding, self.errors))

    async def close(self):
        """
        Close and release the stream.
        """
        await self.base.close()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


async def async_copyfile_str(copyfrom: AReadText, copyto: AWriteText, blocksize=1024 * 1024):
    """
    Stream text from ``copyfrom`` to ``copyto``.
    """
    while True:
        data = await copyfrom.read(blocksize)
        if not data:
            break
        await copyto.write(data)


async def roundrobin(iterables: List):
    """
    An async version of the itertools roundrobin recipe.

    roundrobin('ABC', 'D', 'EF') --> A D E B F C
    """
    i = 0
    while iterables:
        i %= len(iterables)
        try:
            yield await iterables[i].__anext__()
        except StopAsyncIteration:
            iterables.pop(i)
        else:
            i += 1



"""
A process manager can be used with a ProcessTask to specify where and how the processes should run.
All a process manager needs to specify is how to launch a process and manipulate the filesystem, and the ProcessTask
will set up an appropriate environment for running the task and retrieve the results using this interface.
"""

from typing import (
    TYPE_CHECKING,
    AsyncContextManager,
    Callable,
    Dict,
    List,
    Literal,
    Optional,
    Set,
    Union,
)
from abc import abstractmethod
from pathlib import Path
import asyncio
import os
import shlex
import signal

import aiofiles
import aioshutil
import asyncssh
import psutil

from .task import STDOUT, _StderrIsStdout

if TYPE_CHECKING:
    from .repository import AReadStream, AReadText, AWriteStream, AWriteText

__all__ = ("AbstractProcessManager", "LocalLinuxManager", "SSHLinuxManager")


class AbstractProcessManager:
    """
    The base class for process managers.

    Processes are managed through arbitrary "process identifier" handle strings.
    """

    @abstractmethod
    async def get_live_pids(self, hint: Set[str]) -> Set[str]:
        """
        Get a set of the live process identifiers. This may include processes which are not part of the current
        app/task, but MUST include all the processes in ``hint`` which are still alive.
        """
        raise NotImplementedError

    @abstractmethod
    async def spawn(
        self,
        args: List[str],
        environ: Dict[str, str],
        cwd: str,
        return_code: str,
        stdin: Optional[str],
        stdout: Optional[str],
        stderr: Optional[Union[str, _StderrIsStdout]],
    ) -> str:
        """
        Launch a process on the target system. This function MUST NOT wait until the process has terminated before
        returning.

        :param args:         The command line of the process to launch. ``args[0]`` is the executable to run.
        :param environ:      A set of environment variables to add to the target process' environment.
        :param cwd:          The directory to launch the process in.
        :param return_code:  A filepath in which to store the process exit code, as an ascii integer.
        :param stdin:        A filepath from which to read the process' stdin, or None for a null file.
        :param stdout:       A filepath to which to write the process' stdout, or None for a null file.
        :param stderr:       A filepath to which to write the process' stderr, None for a null file, or the constant
                             `pydatatask.task.STDOUT` to interleave it into the stdout stream.
        :return:             The process identifier of the process spawned.
        """
        raise NotImplementedError

    @abstractmethod
    async def kill(self, pid: str):
        """
        Terminate the process with the given identifier. This should do nothing if the process is not currently running.
        This should not prevent e.g. the ``return_code`` file from being populated.
        """
        raise NotImplementedError

    @property
    @abstractmethod
    def basedir(self) -> Path:
        """
        A path on the target system to a directory which can be freely manipulated by the app.
        """
        raise NotImplementedError

    @abstractmethod
    async def open(
        self, path: Path, mode: Literal["r", "rb", "w", "wb"]
    ) -> Union["AReadText", "AWriteText", "AReadStream", "AWriteStream"]:
        """
        Open the file on the target system for reading or writing according to the given mode.
        """
        raise NotImplementedError

    @abstractmethod
    async def mkdir(self, path: Path):  # exist_ok=True, parents=True
        """
        Create a directory with the given path on the target system.

        This should have the semantics of ``mkdir -p``, i.e. create any necessary parent directories and also succeed if
        the directory already exists.
        """
        raise NotImplementedError

    @abstractmethod
    async def rmtree(self, path: Path):
        """
        Remove a directory and all its children with the given path on the target system.
        """
        raise NotImplementedError


class LocalLinuxManager(AbstractProcessManager):
    """
    A process manager to run tasks on the local linux machine. By default, it will create a directory
    ``/tmp/pydatatask`` in which to store data.
    """

    def __init__(self, app: str, local_path: Union[Path, str] = "/tmp/pydatatask"):
        self.local_path = Path(local_path) / app

    @property
    def basedir(self) -> Path:
        return self.local_path

    async def get_live_pids(self, hint: Set[str]) -> Set[str]:
        return {str(x) for x in psutil.pids()}

    async def spawn(self, args, environ, cwd, return_code, stdin, stdout, stderr):
        if stdin is None:
            stdin = "/dev/null"
        else:
            stdin = shlex.quote(stdin)
        if stdout is None:
            stdout = "/dev/null"
        else:
            stdout = shlex.quote(stdout)
        if stderr is None:
            stderr = "/dev/null"
        elif stderr is STDOUT:
            stderr = "&1"
        else:
            stderr = shlex.quote(stderr)
        if "/" in args[0]:
            os.chmod(args[0], 0o755)
        return_code = shlex.quote(return_code)
        p = await asyncio.create_subprocess_shell(
            f"""
            {shlex.join(args)} <{stdin} >{stdout} 2>{stderr} &
            PID=$!
            echo $PID
            wait $PID >/dev/null 2>/dev/null
            echo $? >{return_code} 2>/dev/null
            """,
            stdout=asyncio.subprocess.PIPE,
            close_fds=True,
            cwd=cwd,
            env=environ,
        )
        pid = await p.stdout.readline()
        return pid.strip()

    async def kill(self, pid: str):
        os.kill(int(pid), signal.SIGKILL)

    async def mkdir(self, path: Path):
        path.mkdir(parents=True, exist_ok=True)

    async def rmtree(self, path: Path):
        await aioshutil.rmtree(path)

    async def open(self, path, mode):
        return aiofiles.open(path, mode)


class SSHLinuxFile:
    """
    A file returned by `SSHLinuxManager.open`. Probably don't instantiate this directly.
    """

    def __init__(self, path: Union[Path, str], mode: Literal["r", "w", "rb", "wb"], ssh: asyncssh.SSHClientConnection):
        self.path = Path(path)
        self.mode = mode
        self.ssh = ssh
        self.sftp_mgr: Optional[AsyncContextManager] = None
        self.sftp: Optional[asyncssh.SFTPClient] = None
        self.fp_mgr: Optional[AsyncContextManager] = None
        self.fp: Optional[asyncssh.SFTPClientFile] = None

    async def __aenter__(self) -> asyncssh.SFTPClientFile:
        self.sftp_mgr = self.ssh.start_sftp_client()
        self.sftp = await self.sftp_mgr.__aenter__()
        self.fp_mgr = self.sftp.open(self.path, self.mode)
        self.fp = await self.fp_mgr.__aenter__()
        return self.fp

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.fp_mgr.__aexit__(exc_type, exc_val, exc_tb)
        await self.sftp_mgr.__aexit__(exc_type, exc_val, exc_tb)


class SSHLinuxManager(AbstractProcessManager):
    """
    A process manager that runs its processes on a remote linux machine, accessed over SSH.
    The SSH connection is parameterized by an :external:class:`asyncssh.connection.SSHClientConnection` instance.

    Sample usage:

    .. code:: python

        session = pydatatask.Session()

        @session.resource
        async def ssh():
            async with asyncssh.connect(
                "localhost", port=self.port, username="weh", password="weh", known_hosts=None
            ) as s:
                yield s

        @session.resource
        async def procman():
            yield pydatatask.SSHLinuxManager(self.test_id, ssh)

        task = pydatatask.ProcessTask("mytask", procman, ...)
    """

    def __init__(
        self,
        app: str,
        ssh: Callable[[], asyncssh.SSHClientConnection],
        remote_path: Union[Path, str] = "/tmp/pydatatask",
    ):
        self.remote_path = Path(remote_path) / app
        self._ssh = ssh

    @property
    def ssh(self) -> asyncssh.SSHClientConnection:
        """
        The `asyncssh.SSHClientConnection` instance associated. Will fail if the connection is provided by an unopened
        Session.
        """
        return self._ssh()

    @property
    def basedir(self) -> Path:
        return self.remote_path

    async def open(self, path, mode):
        return SSHLinuxFile(path, mode, self.ssh)

    async def rmtree(self, path: Path):
        async with self.ssh.start_sftp_client() as sftp:
            await sftp.rmtree(path, ignore_errors=True)

    async def mkdir(self, path: Path):
        async with self.ssh.start_sftp_client() as sftp:
            await sftp.makedirs(path, exist_ok=True)

    async def kill(self, pid: str):
        await self.ssh.run(f"kill -9 {pid}")

    async def get_live_pids(self, hint):
        p = await self.ssh.run("ls /proc")
        return {x for x in p.stdout.split() if x.isdigit()}

    async def spawn(self, args, environ, cwd, return_code, stdin, stdout, stderr):
        if stdin is None:
            stdin = "/dev/null"
        else:
            stdin = shlex.quote(stdin)
        if stdout is None:
            stdout = "/dev/null"
        else:
            stdout = shlex.quote(stdout)
        if stderr is None:
            stderr = "/dev/null"
        elif stderr is STDOUT:
            stderr = "&1"
        else:
            stderr = shlex.quote(stderr)
        if "/" in args[0]:
            async with self.ssh.start_sftp_client() as sftp:
                await sftp.chmod(args[0], 0o755)
        p = await self.ssh.create_process(
            f"""
            cd {shlex.quote(cwd)}
            {shlex.join(args)} <{stdin} >{stdout} 2>{stderr} &
            PID=$!
            echo $PID
            wait $PID >/dev/null 2>/dev/null
            echo $? >{shlex.quote(return_code)} 2>/dev/null
            """,
            env=environ,
        )
        pid = await p.stdout.readline()
        return pid.strip()



"""
pydatatask defines the notion of resources, or numerical quantities of CPU and memory which can be allocated to a given
job. This is mediated through a :class:`ResourceManager`, an object which can atomically track increments and decrements
from a quota, and reject a request if it would break the quota.

Typical usage is to construct a :class:`ResourceManager` and pass it to a task constructor:

.. code:: python

    quota = pydatatask.ResourceManager(pydatatask.Resources.parse(cpu='1000m', mem='1Gi'))
    task = pydatatask.ProcessTask("my_task", localhost, quota, ...)
"""
from typing import Awaitable, Callable, List, Optional, Union
from asyncio import Lock
from dataclasses import dataclass, field
from decimal import Decimal
from enum import Enum, auto

from kubernetes.utils import parse_quantity

__all__ = ("ResourceType", "Resources", "ResourceManager", "parse_quantity")


class ResourceType(Enum):
    """
    An enum class indicating a type of resource. Presently can be CPU or MEM.
    """

    CPU = auto()
    MEM = auto()


@dataclass
class Resources:
    """
    A dataclass containing a quantity of resources.

    Resources can be summed:

    .. code:: python

        r = pydatatask.Resources.parse(1, 1)
        r += pydatatask.Resources.parse(2, 3)
        r -= pydatatask.Resources.parse(1, 1)
        assert r == pydatatask.Resources.parse(2, 3)
    """

    cpu: Decimal = field(default=Decimal(0))
    mem: Decimal = field(default=Decimal(0))

    @staticmethod
    def parse(cpu: Union[str, float, int, Decimal], mem: Union[str, float, int, Decimal]) -> "Resources":
        """
        Construct a :class:`Resources` instance by parsing the given quantities of CPU and memory.
        """
        return Resources(cpu=parse_quantity(cpu), mem=parse_quantity(mem))

    def __add__(self, other: "Resources"):
        return Resources(cpu=self.cpu + other.cpu, mem=self.mem + other.mem)

    def __mul__(self, other: int):
        return Resources(cpu=self.cpu * other, mem=self.mem * other)

    def __sub__(self, other: "Resources"):
        return self + other * -1

    def excess(self, limit: "Resources") -> Optional[ResourceType]:
        """
        Determine if these resources are over a given limit.

        :return: The ResourceType of the first resource that is over-limit, or None if self is under-limit.
        """
        if self.cpu > limit.cpu:
            return ResourceType.CPU
        elif self.mem > limit.mem:
            return ResourceType.MEM
        else:
            return None


class ResourceManager:
    """
    The ResourceManager tracks quotas of resources. Direct use of this class beyond construction will only be necessary
    if you are writing a custom Task class.
    """

    def __init__(self, quota: Resources):
        """
        :param quota: The resource limit that should be applied to the sum of all tasks using this manager.
        """
        self.quota = quota
        self._lock = Lock()
        self._cached: Optional[Resources] = None
        self._registered_getters: List[Callable[[], Awaitable[Resources]]] = []

    def register(self, func: Callable[[], Awaitable[Resources]]):
        """
        Register an async callback which will be used to load the current resource utilization on process start.
        The initial reported usage will be the sum of the result of all the registered callbacks.

        If you are writing a Task class, you should call this in your constructor to save a reference to a method on
        your class which retrieves the current resource utilization by that task.
        """
        self._registered_getters.append(func)

    async def _getter(self):
        result = Resources()
        for getter in self._registered_getters:
            result += await getter()
        return result

    async def reserve(self, request: Resources) -> Optional[ResourceType]:
        """
        Atomically reserve the given amount of resources and return None, or do nothing and return the limiting resource
        type if any resource would be over-quota.
        """
        async with self._lock:
            if self._cached is None:
                self._cached = await self._getter()

            target = self._cached + request
            excess = target.excess(self.quota)
            if excess is None:
                self._cached = target
                return None
            else:
                return excess

    async def relinquish(self, request: Resources):
        """
        Atomically mark the given amount of resources as available.
        """
        async with self._lock:
            if self._cached is None:
                self._cached = await self._getter()

            self._cached -= request



"""
Repositories are arbitrary key-value stores. They are the data part of pydatatask.
You can store your data in any way you desire and as long as you can write a Repository class to describe it, it can be
used to drive a pipeline.

The notion of the "value" part of the key-value store abstraction is defined very, very loosely. The repository base
class doesn't have an interface to get or store values, only to query for and delete keys. Instead, you have to know
which repository subclass you're working with, and use its interfaces. For example, `MetadataRepository` assumes that
its values are structured objects and loads them fully into memory, and `BlobRepository` provides a streaming interface
to a flat address space.
"""
from typing import (
    TYPE_CHECKING,
    Any,
    AsyncGenerator,
    AsyncIterable,
    Awaitable,
    Callable,
    Coroutine,
    Dict,
    List,
    Literal,
    Optional,
    overload,
)
from abc import ABC, abstractmethod
from collections import Counter
from pathlib import Path
import base64
import hashlib
import inspect
import io
import logging
import os
import string

from kubernetes_asyncio.client import V1Pod
from types_aiobotocore_s3.client import S3Client
import aiofiles.os
import aiohttp.client_exceptions
import aioshutil
import botocore.exceptions
import docker_registry_client_async
import dxf
import motor.motor_asyncio
import yaml

from .utils import AReadStream, AReadText, AWriteStream, AWriteText, roundrobin

if TYPE_CHECKING:
    from .task import ExecutorTask, KubeTask

l = logging.getLogger(__name__)

__all__ = (
    "Repository",
    "BlobRepository",
    "MetadataRepository",
    "FileRepositoryBase",
    "FileRepository",
    "DirectoryRepository",
    "S3BucketRepository",
    "S3BucketInfo",
    "MongoMetadataRepository",
    "InProcessMetadataRepository",
    "InProcessBlobStream",
    "InProcessBlobRepository",
    "DockerRepository",
    "LiveKubeRepository",
    "ExecutorLiveRepo",
    "AggregateOrRepository",
    "AggregateAndRepository",
    "BlockingRepository",
    "YamlMetadataRepository",
    "YamlMetadataFileRepository",
    "YamlMetadataS3Repository",
    "RelatedItemRepository",
)


def job_getter(f):
    """
    Use this function to annotate non-abstract methods which take a job identifier as their first parameter. This is
    used by RelatedItemRepository to automatically translate job identifiers to related ones.
    """
    if not inspect.iscoroutinefunction(f):
        raise TypeError("only async functions can be job_getters")
    f.is_job_getter = True
    return f


class Repository(ABC):
    """
    A repository is a key-value store where the keys are names of jobs. Since the values have unspecified semantics, the
    only operations you can do on a generic repository are query for keys.

    A repository can be async-iterated to get a listing of its members.
    """

    CHARSET = CHARSET_START_END = string.ascii_letters + string.digits

    @classmethod
    def is_valid_job_id(cls, job: str):
        """
        Determine whether the given job identifier is valid, i.e. that it contains only valid characters
        (numbers and letters by default).
        """
        return (
            0 < len(job) < 64
            and all(c in cls.CHARSET for c in job)
            and job[0] in cls.CHARSET_START_END
            and job[-1] in cls.CHARSET_START_END
        )

    async def filter_jobs(self, iterator: AsyncIterable[str]) -> AsyncIterable[str]:
        """
        Apply `is_valid_job_id` as a filter to an async iterator.
        """
        async for job in iterator:
            if self.is_valid_job_id(job):
                yield job
            else:
                l.warning("Skipping %s %s - not a valid job id", self, repr(job))

    async def contains(self, item):
        """
        Determine whether the given job identifier is present in this repository.

        The default implementation is quite inefficient; please override this if possible.
        """
        async for x in self:
            if x == item:
                return True
        return False

    def __aiter__(self):
        return self.filter_jobs(self.unfiltered_iter())

    @abstractmethod
    async def unfiltered_iter(self) -> AsyncGenerator[str, None]:
        """
        The core method of Repository. Implement this to produce an iterable of every string which could potentially
        be a job identifier present in this repository. When the repository is iterated directly, this will be filtered
        by `filter_jobs`.
        """
        raise NotImplementedError
        # noinspection PyUnreachableCode
        yield None  # pylint: disable=unreachable

    @abstractmethod
    async def info(self, job) -> Any:
        """
        Returns an arbitrary piece of data related to job. Notably, this is used during templating.
        This should do something meaningful even if the repository does not contain the requested job.
        """
        raise NotImplementedError

    @abstractmethod
    async def delete(self, job):
        """
        Delete the given job from the repository. This should succeed even if the job is not present in this repository.
        """
        raise NotImplementedError

    async def info_all(self) -> Dict[str, Any]:
        """
        Produce a mapping from every job present in the repository to its corresponding info. The default implementation
        is somewhat inefficient; please override it if there is a more effective way to load all info.
        """
        return {job: await self.info(job) async for job in self}

    async def validate(self):
        """
        Override this method to raise an exception if for any reason the repository is misconfigured. This will be
        automatically called by the pipeline on opening.
        """

    def map(
        self, func: Callable, filt: Optional[Callable[[str], Awaitable[bool]]] = None, allow_deletes=False
    ) -> "MapRepository":
        """
        Generate a :class:`MapRepository` based on this repository and the given parameters.
        """
        return MapRepository(self, func, filt, allow_deletes=allow_deletes)


class MapRepository(Repository):
    """
    A MapRepository is a repository which uses arbitrary functions to map and filter results from a base repository.
    """

    def __init__(
        self,
        base: Repository,
        func: Callable[[Any], Coroutine[None, None, Any]],
        filt: Optional[Callable[[str], Awaitable[bool]]] = None,
        allow_deletes=False,
    ):
        """
        :param func: The function to use to translate the base repository's `info` results into the mapped `info`
                     results.
        :param filt: Optional: An async function to use to determine whether a given key should be considered part of
                     the mapped repository.
        :param allow_deletes: Whether the delete operation will do anything on the mapped repository.
        """
        self.base = base
        self.func = func
        self.filter = filt
        self.allow_deletes = allow_deletes

    async def contains(self, item):
        if self.filter is None or await self.filter(item):
            return await self.base.contains(item)
        return False

    async def delete(self, job):
        if self.allow_deletes:
            await self.base.delete(job)

    async def unfiltered_iter(self):
        async for item in self.base.unfiltered_iter():
            if self.filter is None or await self.filter(item):
                yield item

    async def info(self, job):
        return await self.func(await self.base.info(job))

    async def info_all(self) -> Dict[str, Any]:
        result = await self.base.info_all()
        to_remove = []
        for k, v in result.items():
            if self.filter is None or await self.filter(k):
                result[k] = await self.func(v)
            else:
                to_remove.append(k)
        for k in to_remove:
            result.pop(k)
        return result


class MetadataRepository(Repository, ABC):
    """
    A metadata repository has values which are small, structured data, and loads them entirely into memory, returning
    the structured data from the `info` method.
    """

    @abstractmethod
    async def info(self, job):
        """
        Retrieve the data with key ``job`` from the repository.
        """
        raise NotImplementedError

    @abstractmethod
    async def dump(self, job, data):
        """
        Insert ``data`` into the repository with key ``job``.
        """
        raise NotImplementedError


class BlobRepository(Repository, ABC):
    """
    A blob repository has values which are flat data blobs that can be streamed for reading or writing.
    """

    @overload
    async def open(self, job: str, mode: Literal["r"]) -> AReadText:
        ...

    @overload
    async def open(self, job: str, mode: Literal["rb"]) -> AReadStream:
        ...

    @overload
    async def open(self, job: str, mode: Literal["w"]) -> AWriteText:
        ...

    @overload
    async def open(self, job: str, mode: Literal["wb"]) -> AWriteStream:
        ...

    @abstractmethod
    async def open(self, job, mode="r"):
        """
        Open the given job's value as a stream for reading or writing, in text or binary mode.
        """
        raise NotImplementedError


class FileRepositoryBase(Repository, ABC):
    """
    A file repository is a local directory where each job identifier is a filename, optionally suffixed with an
    extension before hitting the filesystem. This is an abstract base class for other file repositories which have more
    to say about what is found at these filepaths.
    """

    def __init__(self, basedir, extension="", case_insensitive=False):
        self.basedir = Path(basedir)
        self.extension = extension
        self.case_insensitive = case_insensitive

    async def contains(self, item):
        return await aiofiles.os.path.exists(self.basedir / (item + self.extension))

    def __repr__(self):
        return f'<{type(self).__name__} {self.basedir / ("*" + self.extension)}>'

    async def unfiltered_iter(self):
        for name in await aiofiles.os.listdir(self.basedir):
            if self.case_insensitive:
                cond = name.lower().endswith(self.extension.lower())
            else:
                cond = name.endswith(self.extension)
            if cond:
                yield name[: -len(self.extension) if self.extension else None]

    async def validate(self):
        self.basedir.mkdir(exist_ok=True, parents=True)
        if not os.access(self.basedir, os.W_OK):
            raise PermissionError(f"Cannot write to {self.basedir}")

    def fullpath(self, job) -> Path:
        """
        Construct the full local path of the file corresponding to ``job``.
        """
        return self.basedir / (job + self.extension)

    @job_getter
    async def info(self, job):
        """
        The templating info provided by a file repository is the full path to the corresponding file as a string.
        """
        return str(self.fullpath(job))


class FileRepository(FileRepositoryBase, BlobRepository):
    """
    A file repository whose members are files, treated as streamable blobs.
    """

    @job_getter
    async def open(self, job, mode="r"):
        return aiofiles.open(self.fullpath(job), mode)

    async def delete(self, job):
        try:
            await aiofiles.os.unlink(self.fullpath(job))
        except FileNotFoundError:
            pass


class DirectoryRepository(FileRepositoryBase):
    """
    A file repository whose members are directories.
    """

    def __init__(self, *args, discard_empty=False, **kwargs):
        """
        :param discard_empty: Whether only directories containing at least one member should be considered as "present"
                              in the repository.
        """
        super().__init__(*args, **kwargs)
        self.discard_empty = discard_empty

    @job_getter
    async def mkdir(self, job):
        """
        Create an empty directory corresponding to ``job``. Do nothing if the directory already exists.
        """
        try:
            await aiofiles.os.mkdir(self.fullpath(job))
        except FileExistsError:
            pass

    async def delete(self, job):
        if await self.contains(job):
            await aioshutil.rmtree(self.fullpath(job))

    async def contains(self, item):
        result = await super().contains(item)
        if not self.discard_empty:
            return result
        if not result:
            return False
        return bool(list(await aiofiles.os.listdir(self.fullpath(item))))

    async def unfiltered_iter(self):
        async for item in super().unfiltered_iter():
            if self.discard_empty:
                if bool(list(await aiofiles.os.listdir(self.fullpath(item)))):
                    yield item
            else:
                yield item


class S3BucketBinaryWriter:
    """
    A class for streaming (or buffering) byte data to be written to an `S3BucketRepository`.
    """

    def __init__(self, repo: "S3BucketRepository", job: str):
        self.repo = repo
        self.job = job
        self.buffer = io.BytesIO()
        super().__init__()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()

    async def close(self):
        """
        Close and flush the data to the bucket.
        """
        self.buffer.seek(0, io.SEEK_END)
        size = self.buffer.tell()
        self.buffer.seek(0, io.SEEK_SET)
        await self.repo.client.put_object(
            Bucket=self.repo.bucket,
            Key=self.repo.object_name(self.job),
            Body=self.buffer,
            ContentLength=size,
            ContentType=self.repo.mimetype,
        )

    async def write(self, data: bytes):
        """
        Write some data to the stream.
        """
        self.buffer.write(data)


class S3BucketReader:
    """
    A class for streaming byte data from an `S3BucketRepository`.
    """

    def __init__(self, body):
        self.body = body

    async def close(self):
        """
        Close and release the stream.
        """
        self.body.close()

    async def read(self, n=None):  # pylint: disable=unused-argument :(
        """
        Read the entire body of the blob. Due to API limitations, we can't read less than that at once...
        """
        return await self.body.read()

    async def __aenter__(self):
        await self.body.__aenter__()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.body.__aexit__(exc_type, exc_val, exc_tb)


class S3BucketInfo:
    """
    The data structure returned from :meth:`S3BucketRepository.info`.

    :ivar uri: The s3 URI of the current job's resource, e.g. ``s3://bucket/prefix/job.ext``. ``str(info)`` will also
               return this.
    :ivar endpoint: The URL of the API server providing the S3 interface.
    """

    def __init__(self, endpoint: str, uri: str):
        self.endpoint = endpoint
        self.uri = uri

    def __str__(self):
        return self.uri


class S3BucketRepository(BlobRepository):
    """
    A repository where keys are paths in a S3 bucket. Provides a streaming interface to the corresponding blobs.
    """

    def __init__(
        self,
        client: Callable[[], S3Client],
        bucket: str,
        prefix: str = "",
        extension: str = "",
        mimetype: str = "application/octet-stream",
        incluster_endpoint: Optional[str] = None,
    ):
        """
        :param client: A callable returning an aiobotocore S3 client connected and authenticated to the server you wish
                       to store things on.
        :param bucket: The name of the bucket from which to load and store.
        :param prefix: A prefix to put on the job name before translating it into a bucket path. If this is meant to be
                       a directory name it should end with a slash character.
        :param extension: A suffix to put on the job name before translating it into a bucket path. If this is meant to
                          be a file extension it should start with a dot.
        :param mimetype: The MIME type to set the content when adding data.
        :param incluster_endpoint: Optional: An endpoint URL to provide as the result of info() queries instead of
                                   extracting the URL from ``client``.
        """
        self._client = client
        self.bucket = bucket
        self.prefix = prefix
        self.extension = extension
        self.mimetype = mimetype
        self.incluster_endpoint = incluster_endpoint

    @property
    def client(self):
        """
        The aiobotocore S3 client. This will raise an error if the client comes from a session which is not opened.
        """
        return self._client()

    def __repr__(self):
        return f"<{type(self).__name__} {self.bucket}/{self.prefix}*{self.extension}>"

    async def contains(self, item):
        try:
            await self.client.head_object(Bucket=self.bucket, Key=self.object_name(item))
        except botocore.exceptions.ClientError:
            return False
        else:
            return True

    async def unfiltered_iter(self):
        paginator = self.client.get_paginator("list_objects")
        async for page in paginator.paginate(Bucket=self.bucket, Prefix=self.prefix):
            for obj in page.get("Contents", []):
                if obj["Key"].endswith(self.extension):
                    yield obj["Key"][len(self.prefix) : -len(self.extension) if self.extension else None]

    async def validate(self):
        try:
            await self.client.head_bucket(Bucket=self.bucket)
        except botocore.exceptions.ClientError as e:
            if "404" in str(e):
                await self.client.create_bucket(Bucket=self.bucket)
            else:
                raise

    def object_name(self, job):
        """
        Return the object name for the given job.
        """
        return f"{self.prefix}{job}{self.extension}"

    @job_getter
    async def open(self, job, mode="r"):
        if mode == "wb":
            return S3BucketBinaryWriter(self, job)
        elif mode == "w":
            return AWriteText(S3BucketBinaryWriter(self, job))
        elif mode == "rb":
            return S3BucketReader((await self.client.get_object(Bucket=self.bucket, Key=self.object_name(job)))["Body"])
        elif mode == "r":
            return AReadText(
                S3BucketReader((await self.client.get_object(Bucket=self.bucket, Key=self.object_name(job)))["Body"])
            )
        else:
            raise ValueError(mode)

    @job_getter
    async def info(self, job):
        """
        Return an `S3BucketInfo` corresponding to the given job.
        """
        return S3BucketInfo(
            self.incluster_endpoint or self.client._endpoint.host,
            f"s3://{self.bucket}/{self.object_name(job)}",
        )

    async def delete(self, job):
        await self.client.delete_object(Bucket=self.bucket, Key=self.object_name(job))


class MongoMetadataRepository(MetadataRepository):
    """
    A metadata repository using a MongoDB collection as the backing store.
    """

    def __init__(
        self,
        collection: Callable[[], motor.motor_asyncio.AsyncIOMotorCollection],
        subcollection: Optional[str],
    ):
        """
        :param collection: A callable returning a motor async collection.
        :param subcollection: Optional: the name of a subcollection within the collection in which to store data.
        """
        self._collection = collection
        self._subcollection = subcollection

    def __repr__(self):
        return f"<{type(self).__name__} {self._subcollection}>"

    @property
    def collection(self) -> motor.motor_asyncio.AsyncIOMotorCollection:
        """
        The motor async collection data will be stored in. If this is provided by an unopened session, raise an error.
        """
        result = self._collection()
        if self._subcollection is not None:
            result = result[self._subcollection]
        return result

    async def contains(self, item):
        return await self.collection.count_documents({"_id": item}) != 0

    async def delete(self, job):
        await self.collection.delete_one({"_id": job})

    async def unfiltered_iter(self):
        async for x in self.collection.find({}, projection=[]):
            yield x["_id"]

    @job_getter
    async def info(self, job):
        """
        The info of a mongo metadata repository is the literal value stored in the repository with identifier ``job``.
        """
        result = await self.collection.find_one({"_id": job})
        if result is None:
            result = {}
        return result

    async def info_all(self) -> Dict[str, Any]:
        return {entry["_id"]: entry async for entry in self.collection.find({})}

    @job_getter
    async def dump(self, job, data):
        await self.collection.replace_one({"_id": job}, data, upsert=True)


class DockerRepository(Repository):
    """
    A docker repository is, well, an actual docker repository hosted in some registry somewhere. Keys translate to tags
    on this repository.
    """

    def __init__(
        self,
        registry: Callable[[], docker_registry_client_async.DockerRegistryClientAsync],
        domain: str,
        repository: str,
    ):
        """
        :param registry: A callable returning a
                         `docker_registry_client_async <https://pypi.org/project/docker-registry-client-async/>`_
                         client object with appropriate authentication information.
        :param domain: The registry domain to connect to, e.g. ``index.docker.io``.
        :param repository: The repository to store images in within the domain, e.g. ``myname/myrepo``.
        """
        self._registry = registry
        self.domain = domain
        self.repository = repository

    @property
    def registry(self) -> docker_registry_client_async.DockerRegistryClientAsync:
        """
        The ``docker_registry_client_async`` client object. If this is provided by an unopened session, raise an error.
        """
        return self._registry()

    async def unfiltered_iter(self):
        try:
            image = docker_registry_client_async.ImageName(self.repository, endpoint=self.domain)
            tags = (await self.registry.get_tags(image)).tags["tags"]
            if tags is None:
                return
            for tag in tags:
                yield tag
        except aiohttp.client_exceptions.ClientResponseError as e:
            if e.status == 404:
                raise

    def __repr__(self):
        return f"<DockerRepository {self.domain}/{self.repository}:*>"

    @job_getter
    async def info(self, job):
        """
        The info provided by a docker repository is a dict with two keys, "withdomain" and "withoutdomain". e.g.:

        .. code::

            { "withdomain": "docker.example.com/myname/myrepo:job", "withoutdomain": "myname/myrepo:job" }
        """
        return {
            "withdomain": f"{self.domain}/{self.repository}:{job}",
            "withoutdomain": f"{self.repository}:{job}",
        }

    def _dxf_auth(self, dxf_obj, response):
        # what a fucking hack
        for pattern, credentials in self.registry.credentials.items():
            if pattern.fullmatch(self.domain):
                result = credentials
                break
        else:
            raise PermissionError("Missing credentials for %s" % self.domain)
        if self.registry.ssl:
            username, password = base64.b64decode(result).decode().split(":")
            dxf_obj.authenticate(username, password, response)
        else:
            dxf_obj._headers = {"Authorization": "Basic " + result}

    async def delete(self, job):
        if not await self.contains(job):
            return

        self._delete_inner(job)  # blocking! epic fail

    def _delete_inner(self, job):
        random_data = os.urandom(16)
        random_digest = "sha256:" + hashlib.sha256(random_data).hexdigest()

        d = dxf.DXF(
            host=self.domain,
            repo=self.repository,
            auth=self._dxf_auth,
            insecure=not self.registry.ssl,
        )
        d.push_blob(data=random_data, digest=random_digest)
        d.set_alias(job, random_digest)
        d.del_alias(job)


class LiveKubeRepository(Repository):
    """
    A repository where keys translate to ``job`` labels on running kube pods. This repository is constructed
    automatically by a `KubeTask` or subclass and is linked as the ``live`` repository. Do not construct this class
    manually.
    """

    def __init__(self, task: "KubeTask"):
        self.task = task

    async def unfiltered_iter(self):
        for pod in await self.pods():
            yield pod.metadata.labels["job"]

    async def contains(self, item):
        return bool(await self.task.podman.query(task=self.task.name, job=item))

    def __repr__(self):
        return f"<LiveKubeRepository task={self.task.name}>"

    @job_getter
    async def info(self, job):
        """
        Cannot template with live kube info. Implement this if you have something in mind.
        """
        return None

    async def pods(self) -> List[V1Pod]:
        """
        A list of live pod objects corresponding to this repository.
        """
        return await self.task.podman.query(task=self.task.name)

    async def delete(self, job):
        """
        Deleting a job from this repository will delete the pod.
        """
        pods = await self.task.podman.query(job=job, task=self.task.name)
        for pod in pods:  # there... really should be only one
            await self.task.delete(pod)
        # while await self.task.podman.query(job=job, task=self.task.name):
        #    await asyncio.sleep(0.2)


class AggregateAndRepository(Repository):
    """
    A repository which is said to contain a job if all its children also contain that job
    """

    def __init__(self, **children: Repository):
        assert children
        self.children = children

    async def unfiltered_iter(self):
        counting = Counter()
        async for item in roundrobin([child.unfiltered_iter() for child in self.children.values()]):
            counting[item] += 1
            if counting[item] == len(self.children):
                yield item

    async def contains(self, item):
        for child in self.children.values():
            if not await child.contains(item):
                return False
        return True

    @job_getter
    async def info(self, job):
        """
        The info provided by an aggregate And repository is a dict mapping each child's name to that child's info.
        """
        return {name: await child.info(job) for name, child in self.children.items()}

    async def delete(self, job):
        """
        Deleting a job from an aggregate And repository deletes the job from all of its children.
        """
        for child in self.children.values():
            await child.delete(job)


class AggregateOrRepository(Repository):
    """
    A repository which is said to contain a job if any of its children also contain that job
    """

    def __init__(self, **children: Repository):
        assert children
        self.children = children

    async def unfiltered_iter(self):
        seen = set()
        for child in self.children.values():
            async for item in child.unfiltered_iter():
                if item in seen:
                    continue
                seen.add(item)
                yield item

    async def contains(self, item):
        for child in self.children.values():
            if await child.contains(item):
                return True
        return False

    @job_getter
    async def info(self, job):
        """
        The info provided by an aggregate Or repository is a dict mapping each child's name to that child's info.
        """
        return {name: await child.info(job) for name, child in self.children.items()}

    async def delete(self, job):
        """
        Deleting a job from an aggregate Or repository deletes the job from all of its children.
        """
        for child in self.children.values():
            await child.delete(job)


class BlockingRepository(Repository):
    """
    A class that is said to contain a job if ``source`` contains it and ``unless`` does not contain it
    """

    def __init__(self, source: Repository, unless: Repository, enumerate_unless=True):
        self.source = source
        self.unless = unless
        self.enumerate_unless = enumerate_unless

    async def unfiltered_iter(self):
        if self.enumerate_unless:
            blocked = set()
            async for x in self.unless.unfiltered_iter():
                blocked.add(x)
        else:
            blocked = None
        async for item in self.source.unfiltered_iter():
            if self.enumerate_unless and item in blocked:
                continue
            if not self.enumerate_unless and self.unless.contains(item):
                continue
            yield item

    async def contains(self, item):
        return await self.source.contains(item) and not await self.unless.contains(item)

    @job_getter
    async def info(self, job):
        return await self.source.info(job)

    async def delete(self, job):
        await self.source.delete(job)


class YamlMetadataRepository(BlobRepository, MetadataRepository, ABC):
    """
    A metadata repository based on a blob repository. When info is accessed, it will **load the target file into
    memory**, parse it as yaml, and return the resulting object.

    This is a base class, and must be overridden to implement the blob loading portion.
    """

    @job_getter
    async def info(self, job):
        async with await self.open(job, "rb") as fp:
            s = await fp.read()
        return yaml.safe_load(s)

    @job_getter
    async def dump(self, job, data):
        s = yaml.safe_dump(data, None)
        async with await self.open(job, "w") as fp:
            await fp.write(s)


class YamlMetadataFileRepository(YamlMetadataRepository, FileRepository):
    """
    A metadata repository based on a file blob repository.
    """

    def __init__(self, filename, extension=".yaml", case_insensitive=False):
        super().__init__(filename, extension=extension, case_insensitive=case_insensitive)


class YamlMetadataS3Repository(YamlMetadataRepository, S3BucketRepository):
    """
    A metadata repository based on a s3 bucket repository.
    """

    def __init__(self, client, bucket, prefix, extension=".yaml", mimetype="text/yaml"):
        super().__init__(client, bucket, prefix, extension=extension, mimetype=mimetype)

    @job_getter
    async def info(self, job):
        try:
            return await super().info(job)
        except botocore.exceptions.ClientError as e:
            if "NoSuchKey" in str(e):
                return {}
            else:
                raise


class RelatedItemRepository(Repository):
    """
    A repository which returns items from another repository based on following a related-item lookup.
    """

    def __init__(
        self,
        base_repository: Repository,
        translator_repository: Repository,
        allow_deletes=False,
        prefetch_lookup=True,
    ):
        """
        :param base_repository: The repository from which to return results based on translated keys. The resulting
                                repository will duck-type as the same type as the base.
        :param translator_repository: A repository whose info() will be used to translate keys:
                                      ``info(job) == translated_job``.
        :param allow_deletes: Whether the delete operation on this repository does anything. If enabled, it will delete
                              only from the base repository.
        :param prefetch_lookup: Whether to cache the entirety of the translator repository in memory to improve
                                performance.
        """
        self.base_repository = base_repository
        self.translator_repository = translator_repository
        self.allow_deletes = allow_deletes
        self.prefetch_lookup_setting = prefetch_lookup
        self.prefetch_lookup = None

    def __repr__(self):
        return f"<{type(self).__name__} {self.base_repository} by {self.translator_repository}>"

    async def _lookup(self, item):
        if self.prefetch_lookup is None and self.prefetch_lookup_setting:
            self.prefetch_lookup = await self.translator_repository.info_all()
        if self.prefetch_lookup:
            return self.prefetch_lookup.get(item)
        else:
            return await self.translator_repository.info(item)

    async def contains(self, item):
        basename = await self._lookup(item)
        if basename is None:
            return False
        return await self.base_repository.contains(basename)

    async def delete(self, job):
        if not self.allow_deletes:
            return

        basename = await self._lookup(job)
        if basename is None:
            return

        await self.base_repository.delete(basename)

    @job_getter
    async def info(self, job):
        basename = await self._lookup(job)
        if basename is None:
            raise LookupError(job)

        return await self.base_repository.info(basename)

    def __getattr__(self, item):
        v = getattr(self.base_repository, item)
        if not getattr(v, "is_job_getter", False):
            return v

        async def inner(job, *args, **kwargs):
            basename = await self._lookup(job)
            if basename is None:
                raise LookupError(job)
            return await v(basename, *args, **kwargs)

        return inner

    async def unfiltered_iter(self):
        base_contents = {x async for x in self.base_repository}
        async for item in self.translator_repository:
            basename = await self._lookup(item)
            if basename is not None and basename in base_contents:
                yield item


class ExecutorLiveRepo(Repository):
    """
    A repository where keys translate to running jobs in an ExecutorTask. This repository is constructed automatically
    and is linked as the ``live`` repository. Do not construct this class manually.
    """

    def __init__(self, task: "ExecutorTask"):
        self.task = task

    def __repr__(self):
        return f"<{type(self).__name__} task={self.task.name}>"

    async def unfiltered_iter(self):
        for job in self.task.rev_jobs:
            yield job

    async def contains(self, item):
        return item in self.task.rev_jobs

    async def delete(self, job):
        """
        Deleting a job from the repository will cancel the corresponding task.
        """
        await self.task.cancel(job)

    async def info(self, job):
        """
        There is no templating info for an `ExecutorLiveRepo`.
        """
        return None


class InProcessMetadataRepository(MetadataRepository):
    """
    An incredibly simple metadata repository which stores all its values in a dict, and will let them vanish when the
    process terminates.
    """

    def __init__(self, data: Optional[Dict[str, Any]] = None):
        self.data: Dict[str, Any] = data if data is not None else {}

    def __repr__(self):
        return f"<{type(self).__name__}>"

    @job_getter
    async def info(self, job):
        return self.data.get(job)

    @job_getter
    async def dump(self, job, data):
        self.data[job] = data

    async def contains(self, item):
        return item in self.data

    async def delete(self, job):
        del self.data[job]

    async def unfiltered_iter(self):
        for job in self.data:
            yield job


class InProcessBlobStream:
    """
    A stream returned from an `BlobRepository.open` call from `InProcessBlobRepository`. Do not construct this manually.
    """

    def __init__(self, repo: "InProcessBlobRepository", job: str):  # pylint: disable=missing-function-docstring
        self.repo = repo
        self.job = job
        self.data = io.BytesIO(repo.data.get(job, b""))

    async def read(self, n: Optional[int] = None) -> bytes:
        """
        Read up to ``n`` bytes from the stream.
        """
        return self.data.read(n)

    async def write(self, data: bytes):
        """
        Write ``data`` to the stream.
        """
        self.data.write(data)

    async def close(self):
        """
        Close and release the stream, syncing the data back to the repository.
        """
        self.repo.data[self.job] = self.data.getvalue()

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.close()


class InProcessBlobRepository(BlobRepository):
    """
    An incredibly simple blob repository which stores all its values in a dict, and will let them vanish when the
    process terminates.
    """

    def __init__(self, data: Optional[Dict[str, bytes]] = None):
        self.data = data if data is not None else {}

    def __repr__(self):
        return f"<{type(self).__name__}>"

    @job_getter
    async def info(self, job):
        """
        There is no templating info for an `InProcessBlobRepository`.
        """
        return None

    @job_getter
    async def open(self, job, mode="r"):
        stream = InProcessBlobStream(self, job)
        if mode == "r":
            return AReadText(stream)
        elif mode == "w":
            return AWriteText(stream)
        else:
            return stream

    async def unfiltered_iter(self):
        for item in self.data:
            yield item

    async def contains(self, item):
        return item in self.data

    async def delete(self, job):
        del self.data[job]



"""
The top-level script you write using pydatatask should call `pydatatask.main.main` in its ``if __name__ == '__main__'``
block. This will parse ``sys.argv`` and display the administration interface for the pipeline.

The help screen should look something like this:

.. code::

    $ python3 main.py --help
      usage: main.py [-h] {update,run,status,trace,rm,ls,cat,inject,launch,shell} ...

    positional arguments:
      {update,run,status,trace,rm,ls,cat,inject,launch,shell}
        update              Keep the pipeline in motion
        run                 Run update in a loop until everything is quiet
        status              View the pipeline status
        trace               Track a job's progress through the pipeline
        rm                  Delete data from the pipeline
        ls                  List jobs in a repository
        cat                 Print data from a repository
        inject              Dump data into a repository
        launch              Manually start a task
        shell               Launch an interactive shell to interrogate the pipeline

    options:
      -h, --help            show this help message and exit
"""

from typing import Callable, Dict, List, Optional, Set, Union
import argparse
import asyncio
import logging
import re
import sys

import IPython
import yaml

from .pipeline import Pipeline
from .repository import BlobRepository, MetadataRepository, Repository
from .task import Task

l = logging.getLogger(__name__)
token_re = re.compile(r"\w+\.\w+")

__all__ = (
    "main",
    "update",
    "cat_data",
    "list_data",
    "delete_data",
    "inject_data",
    "print_status",
    "print_trace",
    "launch",
    "shell",
    "run",
)

# pylint: disable=missing-function-docstring,missing-class-docstring


def main(
    pipeline: Pipeline,
    instrument: Optional[Callable[[argparse._SubParsersAction], None]] = None,
):
    """
    The pydatatask main function! Call this with the pipeline you've constructed to parse ``sys.argv`` and display the
    pipeline administration interface.

    If you like, you can pass as the ``instrument`` argument a function which will add additional commands to the menu.
    """
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(dest=argparse.SUPPRESS, required=True)

    parser_update = subparsers.add_parser("update", help="Keep the pipeline in motion")
    parser_update.set_defaults(func=update)

    parser_run = subparsers.add_parser("run", help="Run update in a loop until everything is quiet")
    parser_run.add_argument("--forever", action="store_true", help="Run forever")
    parser_run.add_argument("--launch-once", action="store_true", help="Only evaluates tasks-to-launch once")
    parser_run.set_defaults(func=run)
    parser_run.set_defaults(timeout=None)

    parser_status = subparsers.add_parser("status", help="View the pipeline status")
    parser_status.add_argument(
        "--all",
        "-a",
        dest="all_repos",
        action="store_true",
        help="Show internal repositories",
    )
    parser_status.set_defaults(func=print_status)

    parser_trace = subparsers.add_parser("trace", help="Track a job's progress through the pipeline")
    parser_trace.add_argument(
        "--all",
        "-a",
        dest="all_repos",
        action="store_true",
        help="Show internal repositories",
    )
    parser_trace.add_argument("job", nargs="+", help="Name of job to trace")
    parser_trace.set_defaults(func=print_trace)

    parser_delete = subparsers.add_parser("rm", help="Delete data from the pipeline")
    parser_delete.add_argument("--recursive", "-r", action="store_true", help="Delete dependant data too")
    parser_delete.add_argument(
        "data",
        type=str,
        help="Name of repository [task.repo] or task from which to delete data",
    )
    parser_delete.add_argument("job", type=str, nargs="+", help="Name of job of which to delete data")
    parser_delete.set_defaults(func=delete_data)

    parser_ls = subparsers.add_parser("ls", help="List jobs in a repository")
    parser_ls.add_argument(
        "data",
        type=str,
        nargs="+",
        help="Name of repository [task.repo] from which to list data",
    )
    parser_ls.set_defaults(func=list_data)

    parser_cat = subparsers.add_parser("cat", help="Print data from a repository")
    parser_cat.add_argument("data", type=str, help="Name of repository [task.repo] from which to print data")
    parser_cat.add_argument("job", type=str, help="Name of job of which to delete data")
    parser_cat.set_defaults(func=cat_data)

    parser_inject = subparsers.add_parser("inject", help="Dump data into a repository")
    parser_inject.add_argument("data", type=str, help="Name of repository [task.repo] to which to inject data")
    parser_inject.add_argument("job", type=str, help="Name of job of which to inject data")
    parser_inject.set_defaults(func=inject_data)

    parser_launch = subparsers.add_parser("launch", help="Manually start a task")
    parser_launch.add_argument(dest="task_name", type=str, help="Name of task to launch")
    parser_launch.add_argument("job", type=str, help="Name of job to launch task on")
    parser_launch.add_argument(
        "--force",
        "-f",
        action="store_true",
        help="Launch even if start is inhibited by data",
    )
    parser_launch.add_argument("--sync", action="store_true", help="Run the task in-process, if possible")
    parser_launch.add_argument(
        "--meta",
        action=argparse.BooleanOptionalAction,
        default=True,
        help="Store metadata related to task completion",
    )
    parser_launch.set_defaults(func=launch)

    parser_shell = subparsers.add_parser("shell", help="Launch an interactive shell to interrogate the pipeline")
    parser_shell.set_defaults(func=shell)

    if instrument is not None:
        instrument(subparsers)

    args = parser.parse_args()
    ns = vars(args)
    func = ns.pop("func")
    result_or_coro = func(pipeline, **ns)
    if asyncio.iscoroutine(result_or_coro):
        return asyncio.run(main_inner(pipeline, result_or_coro))
    else:
        return result_or_coro


async def main_inner(pipeline, coro):
    async with pipeline:
        await coro


def shell(pipeline: Pipeline):
    pydatatask = __import__("pydatatask")
    assert pipeline
    assert pydatatask
    IPython.embed(using="asyncio")


async def update(pipeline: Pipeline):
    await pipeline.update()


async def run(pipeline: Pipeline, forever: bool, launch_once: bool, timeout: Optional[float]):
    func = pipeline.update
    start = asyncio.get_running_loop().time()
    while await func() or forever:
        if launch_once:
            func = pipeline.update_only_update
        await asyncio.sleep(1)
        if timeout is not None and asyncio.get_running_loop().time() - start > timeout:
            raise TimeoutError("Pipeline run timeout")


def get_repos(pipeline: Pipeline, all_repos: bool):
    seen = set()
    for task in pipeline.tasks.values():
        for link in task.links.values():
            if not all_repos and not link.is_status and not link.is_input and not link.is_output:
                continue
            if link.repo in seen:
                continue
            seen.add(link.repo)
            yield link.repo


async def print_status(pipeline: Pipeline, all_repos: bool):
    async def inner(repo: Repository):
        the_sum = 0
        async for _ in repo:
            the_sum += 1
        return the_sum

    repo_list = list(get_repos(pipeline, all_repos))
    repo_sizes = dict(zip(repo_list, await asyncio.gather(*(inner(repo) for repo in repo_list))))

    for task in pipeline.tasks.values():
        print(task.name)
        for link_name, link in sorted(
            task.links.items(),
            key=lambda x: 0 if x[1].is_input else 1 if x[1].is_status else 2,
        ):
            if link.repo in repo_sizes:
                print(f"{task.name}.{link_name} {repo_sizes[link.repo]}")
        print()


async def print_trace(pipeline: Pipeline, all_repos: bool, job: List[str]):
    async def inner(repo: Repository):
        return [j for j in job if await repo.contains(j)]

    repo_list = list(get_repos(pipeline, all_repos))
    repo_members = dict(zip(repo_list, await asyncio.gather(*(inner(repo) for repo in repo_list))))

    for task in pipeline.tasks.values():
        print(task.name)
        for link_name, link in sorted(
            task.links.items(),
            key=lambda x: 0 if x[1].is_input else 1 if x[1].is_status else 2,
        ):
            if link.repo in repo_members:
                print(f"{task.name}.{link_name} {' '.join(repo_members[link.repo])}")
        print()


async def delete_data(pipeline: Pipeline, data: str, recursive: bool, job: List[str]):
    item: Union[Task, Repository]
    if "." in data:
        taskname, reponame = data.split(".")
        item = pipeline.tasks[taskname].links[reponame].repo
    else:
        item = pipeline.tasks[data]

    for dependant in pipeline.dependants(item, recursive):
        if isinstance(dependant, Repository):
            if job[0] == "__all__":
                jobs = [x async for x in dependant]
                check = False
            else:
                jobs = job
                check = True

            async def del_job(j, dep):
                await dep.delete(j)
                print(j, dep)

            await asyncio.gather(*[del_job(j, dependant) for j in jobs if not check or await dependant.contains(j)])


async def list_data(pipeline: Pipeline, data: List[str]):
    input_text = " ".join(data)
    tokens = token_re.findall(input_text)
    namespace = {name.replace("-", "_"): TaskNamespace(task) for name, task in pipeline.tasks.items()}
    for token in tokens:
        task, repo = token.split(".")
        await namespace[task].consume(repo)
    result = eval(input_text, {}, namespace)  # pylint: disable=eval-used

    for job in sorted(result):
        print(job)


class TaskNamespace:
    def __init__(self, task: Task):
        self.task = task
        self.repos: Dict[str, Set[str]] = {}

    def __getattr__(self, item):
        return self.repos[item]

    async def consume(self, repo: str):
        result = set()
        async for x in self.task.links[repo].repo:
            result.add(x)
        self.repos[repo] = result


async def cat_data(pipeline: Pipeline, data: str, job: str):
    taskname, reponame = data.split(".")
    item = pipeline.tasks[taskname].links[reponame].repo

    if isinstance(item, BlobRepository):
        async with await item.open(job, "rb") as fp:
            while True:
                data_bytes = await fp.read(4096)
                if not data_bytes:
                    break
                await asyncio.get_running_loop().run_in_executor(None, sys.stdout.buffer.write, data_bytes)
    elif isinstance(item, MetadataRepository):
        data_bytes = await item.info(job)
        data_str = yaml.safe_dump(data_bytes, None)
        await asyncio.get_running_loop().run_in_executor(None, sys.stdout.write, data_str)
    else:
        print("Error: cannot cat a repository which is not a blob or metadata")
        return 1


async def inject_data(pipeline: Pipeline, data: str, job: str):
    taskname, reponame = data.split(".")
    item = pipeline.tasks[taskname].links[reponame].repo

    if isinstance(item, BlobRepository):
        async with await item.open(job, "wb") as fp:
            while True:
                data_bytes = await asyncio.get_running_loop().run_in_executor(None, sys.stdin.buffer.read, 1024 * 1024)
                if not data:
                    break
                await fp.write(data_bytes)
    elif isinstance(item, MetadataRepository):
        try:
            data_obj = yaml.safe_load(sys.stdin)
        except yaml.YAMLError:
            print("Error: could not parse stdin as yaml")
            return 1
        await item.dump(job, data_obj)
    else:
        print("Error: cannot inject data into a repository which is not a blob or metadata")
        return 1


async def launch(pipeline: Pipeline, task_name: str, job: str, sync: bool, meta: bool, force: bool):
    task = pipeline.tasks[task_name]
    pipeline.settings(sync, meta)

    if force or await task.ready.contains(job):
        await task.launch(job)
    else:
        l.warning("Task is not ready to launch - use -f to force")
        return 1



"""
A pipeline is just an unordered collection of tasks. Relationships between the tasks are implicit, defined by which
repositories they share.
"""

from typing import Dict, Iterable, Union
import asyncio
import logging

import networkx.algorithms.traversal.depth_first_search
import networkx.classes.digraph

from .repository import Repository
from .session import Session
from .task import Task

l = logging.getLogger(__name__)

__all__ = ("Pipeline",)


class Pipeline:
    """
    The pipeline class.
    """

    def __init__(self, tasks: Iterable[Task], session: Session):
        """
        :param tasks: The tasks which make up this pipeline.
        :param session: The session to open while this pipeline is active.
        """
        self.tasks: Dict[str, Task] = {task.name: task for task in tasks}
        self._opened = False
        self.session = session

    def settings(self, synchronous=False, metadata=True):
        """
        This method can be called to set properties of the current run.

        :param synchronous: Whether jobs will be started and completed in-process, waiting for their completion before
                            a launch phase succeeds.
        :param metadata: Whether jobs will store their completion metadata.
        """
        for task in self.tasks.values():
            task.synchronous = synchronous
            task.metadata = metadata

    async def _validate(self):
        seen_repos = set()
        for task in self.tasks.values():
            await task.validate()
            for link in task.links.values():
                repo = link.repo
                if repo not in seen_repos:
                    seen_repos.add(repo)
                    await repo.validate()

    async def __aenter__(self):
        await self.open()

    async def open(self):
        """
        Opens the pipeline and its associated resource session. This will be automatically when entering an
        ``async with pipeline:`` block.
        """
        if self._opened:
            raise Exception("Pipeline is alredy used")
        self._opened = True
        await self.session.open()
        await self._validate()

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if exc_val is not None:
            l.error("Terminated with error", exc_info=exc_val)
        await self.close()

    async def close(self):
        """
        Closes the pipeline and its associated resource session. This will be automatically called when exiting an
        ``async with pipeline:`` block.
        """
        await self.session.close()

    async def update(self) -> bool:
        """
        Perform one round of pipeline maintenance, running the update phase and then the launch phase. The pipeline must
        be opened for this function to run.

        :return: Whether there is any activity in the pipeline.
        """
        if not self._opened:
            raise Exception("Pipeline must be opened")

        l.info("Running update...")
        result = await self.update_only_update() | await self.update_only_launch()
        l.debug("Completed update")
        return result

    async def update_only_update(self) -> bool:
        """
        Perform one round of the update phase of pipeline maintenance. The pipeline must be opened for this function to
        run.

        :return: Whether there were any live jobs.
        """
        if not self._opened:
            raise Exception("Pipeline must be opened")

        to_gather = [task.update() for task in self.tasks.values()]
        gathered = await asyncio.gather(*to_gather, return_exceptions=False)
        return any(gathered)

    async def update_only_launch(self):
        """
        Perform one round of the launch phase of pipeline maintenance. The pipeline must be opened for this function to
        run.

        :return: Whether there were any jobs launched or ready to launch.
        """
        if not self._opened:
            raise Exception("Pipeline must be opened")

        to_gather = [task.launch_all() for task in self.tasks.values()]
        gathered = await asyncio.gather(*to_gather, return_exceptions=False)
        return any(gathered)

    def graph(self) -> "networkx.classes.digraph.DiGraph":
        """
        Generate the dependency graph for a pipeline. This is a directed graph containing both repositories and tasks
        as nodes.
        """
        result = networkx.classes.digraph.DiGraph()
        for task in self.tasks.values():
            result.add_node(task)
            for link_name, link in task.links.items():
                attrs = dict(vars(link))
                attrs["link_name"] = link_name
                repo = attrs.pop("repo")
                if attrs["is_input"] or attrs["required_for_start"] or attrs["inhibits_start"]:
                    result.add_edge(repo, task, **attrs)
                    if attrs["is_output"] or attrs["is_status"]:
                        result.add_edge(task, repo, **attrs)
                else:
                    result.add_edge(task, repo, **attrs)

        return result

    def dependants(self, node: Union[Repository, Task], recursive: bool) -> Iterable[Union[Repository, Task]]:
        """
        Iterate the list of repositories that are dependent on the given node of the dependency graph.

        :param node: The starting point of the graph traversal.
        :param recursive: Whether to return only direct dependencies (False) or transitive dependencies (True) of node.
        """
        graph = self.graph()
        if recursive:
            yield from networkx.algorithms.traversal.depth_first_search.dfs_preorder_nodes(graph, source=node)
        else:
            yield node
            yield from graph.successors(node)
